import itertools
import json
import textwrap
import warnings
from os import linesep as nl
from typing import Type, List, Union, Tuple, Dict

from box import Box

from timelink.kleio.groups.kelement import KElement
import logging


class KGroup:
    """
    KGroup(*positional_elements ,**more_elements)

    Abstract Kleio Group.

    To define a Kleio Group extend this class and set default value for _name,
    or use extend(name,position, guaranteed, also).

    Define allowed elements in the default values for _position, _guaranteed,
    _also (list of strings) or call allowed_as_element .

    Use _part to list allowed enclosed groups, or call allow_as_part

    For an example see timelink.kleio.groups.KPerson

    Class Methods:
        - extend: Create a new group definition by extending this one.
        - get_subclasses: Generator for subclasses of this group.
        - all_subclasses: List of all the subclasses of this group.
        - is_kgroup: Check if an object is an instance of a subclass of KGroup.
        - elements_allowed: Get the set of elements allowed in this group.
        - allow_as_element: Add element(s) to the list of allowed elements for this group.
        - allow_as_part: Allow a group to be enclosed as part of this group.
        - inc_sequence: Increment and return the global sequence count.
        - inc_line: Increment and return the global line count.

    Instance Methods:
        - include: Include a group within this group.
        - before_include: Method called before this group is included into another.
        - after_include: Method called after a new group is included in this one.
        - is_allowed_as_part: Test if a group can be included in the current one.
        - is_allowed_as_element: Test if an element is allowed in this group.
        - contains: Returns included groups, optionally filtered by group name.
        - elements: Returns elements in this group.
        - element_names: Returns the names of elements in this group
        - attr: Utility function to include a KAttribute in this KGroup.
        - rel: Include a relation in this KGroup.
        - to_kleio: Return a kleio representation of the group.
        - to_dict: Return group information as a dictionary.
        - to_json: Return group information as a JSON string.
        - to_dots: Return group information as a Box object for easy referencing.
        - __str__: Return a string representation of the group.
        - __getitem__: Get an element by name.
        - __setitem__: Set an element by name.
        - pack_as_kelement: Pack a value as a KElement with a given name.
        - unpack_from_kelement: Unpack the core value from a KElement.
        - get_core: Get the core value of an element.
        - get_id: Return the id of the group.
        - get_element_by_name_or_class: Return the value of an element by name or class.


    """

    # Class scoped list of reserved element names for system use
    _builtin_elements = [
        "line",
        "level",
        "order",
        "inside",
        "groupname",
        "pom_class_id",
        "id",
    ]

    # These are shared between all instances of the class
    _position: list = []  # list of positional elements
    _guaranteed: list = []  # list of required elements
    _also: list = []  # list of optional elements
    _part: list = []  # allowed sub groups
    _extends: Type["KGroup"]  # name of group this is based on
    _pom_class_id: str = "entity"  # Id of PomSom mapper for this group
    _element_check = True  # if true validates element assignment
    _elementsd: Type[Dict[str, KElement]]  # Current elements
    _inside: Type["KGroup"]  # group that includes this

    # The following fields are generated during the translation process.
    # If the Kleio groups are generated programatically (for instance when
    # importing from csv files or from databases) then the cls.include method
    # will update these values
    _line: int  # line in the source file
    _level: int  # nesting always +1  than enclosing group
    _order: int  # sequential number of this group in the original source
    # class scoped counters to ensure proper numbering across different groups
    _global_sequence: int = 1  # global sequence count
    _global_line: int = 1  # global line count

    # we need the classes for attributes and relations, used by the
    # attr and rel methods. Since those class are subclasses of this one
    # they need to set this latter. Also allows for differet classes to
    # be used by attr and rel if necessary
    _attr_class = None  # class used by the attr method (set by KAttribute)
    _rel_class = None  # class used by the rel method (set by KRelation)

    # TODO to_kleio_str generates the definition of a group
    #       for a kleio str file. recurse=yes
    #       collects included groups and generates for those also.

    @property
    def kname(self):
        """The kleio name of this group, used in the to_kleio() method"""
        kname = self.unpack_from_kelement(self._name)
        if kname is None or kname == "kgroup":
            kname = self.__class__._name
        return kname

    @kname.setter
    def kname(self, value):
        self._name = self.pack_as_kelement("kname", value, element_class="name")

    @property
    def inside(self):
        return self._inside

    def get_container_id(self):
        if self._inside is None:
            gid = None
        elif KGroup.is_kgroup(self._inside):
            gid = self.unpack_from_kelement(self.inside.id)
        return gid

    @inside.setter
    def inside(self, value):
        if KGroup.is_kgroup(value):
            self._inside = value
        else:
            self._inside = KGroup(id=str(value), element_check=False)

    @property
    def line(self):
        return self.unpack_from_kelement(self._line)

    @line.setter
    def line(self, value):
        self._line = self.pack_as_kelement("line", value)

    @property
    def level(self):
        return self.unpack_from_kelement(self._level)

    @level.setter
    def level(self, value):
        self._level = self.pack_as_kelement("level", value)

    @property
    def order(self):
        return self.unpack_from_kelement(self._order)

    @order.setter
    def order(self, value):
        self._order = self.pack_as_kelement("order", value)

    @property
    def pom_class_id(self):
        return self.unpack_from_kelement(self._pom_class_id)

    @pom_class_id.setter
    def pom_class_id(self, pcid):
        self._pom_class_id = self.pack_as_kelement("pom_class_id", pcid)

    @classmethod
    def extend(
        cls,
        name: str,
        position: Union[list, str, None] = None,
        guaranteed: Union[list, str, None] = None,
        also: Union[list, str, None] = None,
        part: Union[list, str, None] = None,
        synonyms: Union[Tuple[str, str], None] = None,
    ):
        """
        Create a new group definition by extending this one

        :param name:  name of the new group
        :param position: list of positional elements
        :param guaranteed: list of required elements
        :param also:    list of optional elements
        :param part: list of groups that can be included in this one
        :param synonyms: list of tuples defining synonym for elements
                         e,g, [("ano","year"),...]
        :return: new group class



        """
        trace = f"{cls}.extend({name})"
        logging.debug(trace)
        new_group = type(name, (cls,), {})
        new_group._name = name
        # todo: k,v in kwargs if in cls set if not error
        if type(position) is str:
            position = [position]
        if type(guaranteed) is str:
            guaranteed = [guaranteed]
        if type(also) is str:
            also = [also]
        if position is not None:
            new_group._position = position
        else:
            new_group._position = list(cls._position)
        if guaranteed is not None:
            new_group._guaranteed = guaranteed
        else:
            new_group._guaranteed = list(cls._guaranteed)
        if also is not None:
            new_group._also = also
        else:
            new_group._also = list(cls._also)
        if part is not None:
            new_group._part = part
        else:
            new_group._part = list(cls._part)
        if synonyms is not None:
            try:
                for word, synonym in synonyms:
                    sym_class = KElement.get_class_for(synonym)
                    if sym_class is None:
                        msg = f"synonym error {synonym} is unkown"
                        raise ValueError(msg)
                    else:
                        # this creates a new KElement class
                        # named "word" that inherits from the KElement
                        # class of synonym
                        sym_class.extend(word)
            except ValueError:
                msg = "synonyms must be list of tuples"
                "e.g. [('ano','date').."
                raise (ValueError(msg))

        new_group._extends = cls

        return new_group

    @classmethod
    def get_subclasses(cls):
        """Generator for subclasses of this group"""
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass

    @classmethod
    def all_subclasses(cls):
        """List of all the subclasses of this group"""
        return list(set(cls.get_subclasses()))

    @classmethod
    def is_kgroup(cls, g):
        """True g is an instance of a subclass of KGroup"""
        return issubclass(type(g), KGroup)

    @classmethod
    def elements_allowed(cls) -> set:
        """Set of  Elements allowed in this Group"""
        return set(cls._guaranteed).union(set(cls._also)).union(cls._position)

    @classmethod
    def allow_as_element(cls, ename: Union[str, List[str]], guaranteed=False, also=True, position=None):
        """
        Add element or list to list of allowed elements for this group.
        Optionally define if element(s) is positional,
        required (guaranteed) or optional
        :param ename:  name of element
        :param guaranteed: if True this element is added to list of required elements.
        :param also: if True this element is optional (default)
        :param position: int, this is a positional element, at this position (0 = first position)
        :return: List of allowed elements
        """
        if type(ename) is List:
            for e in ename:
                cls.allow_as_element(e, guaranteed=guaranteed, position=position, also=also)
            return
        elif type(ename) is str:
            if guaranteed:
                if ename not in cls._guaranteed:
                    cls._guaranteed.append(ename)
            if position is not None:
                if ename in cls._position:
                    cls._position.remove(ename)
                cls._position.insert(position, ename)
            if ename not in cls.elements_allowed():
                cls._also.append(ename)
        else:
            raise TypeError("first argument must be string or list of strings")
        return

    @classmethod
    def allow_as_part(cls, g: Union[str, type]):
        """Allow g to be enclosed as part of this group.

        Arguments:
            g: the name of a group, or a subclass of KGroup.
               A string or class.
        """
        if g not in cls._part:
            cls._part.append(g)

    @classmethod
    def inc_sequence(cls):
        KGroup._global_sequence += 1
        return KGroup._global_sequence

    @classmethod
    def inc_line(cls):
        KGroup._global_line += 1
        return KGroup._global_line

    def __init__(self, *args, **kwargs):
        """Creates a new instance of group with setting the value of elements.

        Args:
            *args: values for positional arguments
            **kwargs: values for optional elements.

        Example:
            ``ks = KSource("s1", type="test", loc="auc", ref="alumni", obs="Nested")``

        Use element_check=False to turn off checking of element names

        """
        self.id: str = None
        self.source_id: str = None
        self._name: str = "kgroup"
        self._containsd: dict = {}
        self.level = 1
        self.line = 1
        self.order = 1
        self._element_check = True
        self._elementsd = {}
        self._inside = None
        self.source_id = None

        if len(args) > len(self._position):
            raise ValueError("Too many positional elements")
        n = 0
        # set the positional arguments according to "_position"
        for arg in args:
            e = self._position[n]
            # setattr(self, e, KElement(e, arg))
            self[e] = arg  # this will go through __setitem__
            n = n + 1
        # keyword arguments must be in one the element lists
        for k, v in kwargs.items():
            if k == "element_check":
                self._element_check = v
            else:
                if self._element_check and not self.is_allowed_as_element(k):
                    raise ValueError(f"Element not allowed: {k}")
                self[k] = v
        # test if the compulsory (guaranteed) elements are present

        # names of the elements present and their super classes
        # we use it to check if required elements are present
        # in the form of a sub class of the require elements
        elements_present = [
            el.name
            for el in list(itertools.chain(*[el.inherits_from() for el in self.elements()]))
            if el.name is not None
        ]

        missing_elements = set(self._guaranteed) - set(elements_present)
        if len(missing_elements) > 0:
            raise TypeError(f"{self.kname}: element(s) {missing_elements} in _guaranteed " f"is missing or with None value")

    def include(self, group: Type["KGroup"]):
        """Include a group. `group`, or its class, must in _part list or
            extend a class in the part list.

        Returns self so it is possible to chain: g.include(g2).include(g3)"""

        allowed = self.is_allowed_as_part(group)
        if allowed is None:
            raise ValueError(f"Group {self.kname} cannot contain {group.kname}")

        group.level = self.level + 1
        group.line = KGroup.inc_line()
        group.order = KGroup.inc_sequence()

        # Hook to before input processing in the group being included
        if hasattr(group, "before_include") and callable(group.before_include):
            if not group.before_include(self):
                raise TypeError(f"{group} includding aborted by " f"group.before_include(self)" f" returning False")

        # new style, dictionary based
        k = self._containsd.keys()
        if allowed in k:
            self._containsd[allowed].append(group)
        else:
            self._containsd[allowed] = [group]
        group.inside = self
        # Hook to after input processing in the group being included
        if hasattr(group, "after_include") and callable(group.after_include):
            group.after_include(self)
        return self

    def before_include(self, container_group):
        """Method called before this group is included into another
        through KGroup.include(KGroup)."""

        if self.id is None:
            if container_group.id is None:
                raise ValueError("A group with no id cannot be included in another " "group also without id")
            gid = f"{container_group.id}-{self.order:02d}-{self.kname[0:3]}"
            self["id"] = gid
            return True

        return True

    def after_include(self, group):
        """Method called after a new group is included in this one
        through KGroup.include(KGroup)."""
        pass

    def is_allowed_as_part(self, group):
        """Test if a group can be included in the current one.

        For a group to be allowed for inclusion one of 3 conditions necessary:
            1. the kname of the group is in self._pars
            2. the type of the group is in self._pars
            3. the type of the group inherits from a type in self._pars

        Return key under which the group is allowed (kname, type or super type)
        Return None if not allowed
        """
        if not self.is_kgroup(group):
            raise TypeError("Argument must be subclass of KGroup")
        if group.kname not in self._part:
            allowed_classes = [c for c in self._part if type(c) is not str]
            super_classes = type(group).mro()
            r = list(set(super_classes).intersection(set(allowed_classes)))
            if len(r) == 0:
                allowed = None
            else:
                allowed = r[0]  # kname?
        else:
            allowed = group.kname

        return allowed

    def is_allowed_as_element(self, element_name):
        """
        Test if this element is allowed in this group.

        For an element to be allowed one the following must be true:

            1. part of KGroup._builtin
            2. part of position list
            3. part of guarenteed list
            4. part of also list

        Note that this function is unaffected by self._element_check

        :param element_name: name of element to check

        :return: True if element allowed False otherwise
        """
        all_elements = self._builtin_elements + self._position + self._guaranteed + self._also
        return element_name in all_elements

    def includes(self, *args, **kargs) -> list:
        """Deprecated. Use contains instead."""
        warnings.warn("includes is deprecated. Use contains instead.", DeprecationWarning, stacklevel=2)
        return self.contains(*args, **kargs)

    def elements(self) -> list:
        """Returns a list of all elements in the group."""
        return list(self._elementsd.values())

    def element_names(self) -> list:
        """Returns a list of all element names in the group."""
        return self._elementsd.keys()

    def contains(self, group: Type[Union[str, Type["KGroup"]]] = None) -> list:
        """Returns included groups.

        Groups are returned by the order in _pars.

        :todo:  this would better be a generator, yield instead of extend.

        :param str group: filter by group name
        """
        if group is not None:
            if group in self._containsd.keys():
                return self._containsd[group]
            else:
                if type(group) is str:
                    gname = group
                elif KGroup.is_kgroup(group):
                    gname = group.kname

                inc_by_part_order = []
                classes_in_contains = [c for c in self._containsd.keys() if hasattr(c, "_name")]
                for class_in_contains in classes_in_contains:
                    if class_in_contains._name == gname:
                        inc_by_part_order.extend(self._containsd[class_in_contains])
                return inc_by_part_order
        else:  # no specific subgroup, we return by pars order
            inc_by_part_order = []
            for p in self._part:
                if p in self._containsd.keys():
                    inc_by_part_order.extend(self._containsd[p])

            return inc_by_part_order

    def attr(
        self,
        the_type: Union[str, KElement, Tuple[str, str, str]],
        value: Union[str, KElement, Tuple[str, str, str]],
        date: Union[str, KElement, Tuple[str, str, str]],
        obs=None,
    ):
        """Utility function to include a KAttribute in this KGroup

        The call::

            KGroup.attr('age','25','2021-08-08',obs='in May')

        is short hand for::

            KGroup.include(KAttr('age','25','2021-08-08',obs='in May'))

        Params google style

        :param str or tuple the_type: core or (core,org,comment)
        :param str or tuple value: core or (core,org,comment)
        :param str date: date as string in Kleio format, or (date,org,comment)
        :param str obs: options observation field

        """
        ka = self._attr_class
        self.include(ka(the_type, value, date=date, obs=obs))
        return self

    def rel(
        self,
        the_type: Union[str, tuple],
        value: Union[str, tuple],
        destname: Union[str, tuple],
        destination: Union[str, tuple],
        date: Union[str, tuple] = None,
        obs: str = None,
    ):
        """include a relation in this KGroup"""
        kr = self._rel_class
        self.include(kr(the_type, value, destname, destination, date=date, obs=obs))

    def to_kleio(self, indent="") -> str:
        """Return a kleio representation of the group."""
        return self.__str__(indent=indent, recurse=True)

    def to_dict(
        self,
        allow_none: bool = False,
        include_str: bool = False,
        include_kleio: bool = False,
        redundant_keys: bool = True,
        include_builtin: bool = True,
    ):
        """Return group information as a dict.

        Params:
            allow_none bool = Include null values (default False)

            :param allow_none: Include null values (default False)
            :param include_str: include a string represention of each element
                                with keys [elementname]_str, the value will include
                                # and % if necessary
            :param include_kleio: include a kleio representation of each element
                                  with keys [elementname]_kleio; this is similar
                                  to *include_str* but the value also contains the
                                  element name
            :param redundant_keys: include redundant keys (see bellow)
            :param include_builtin: include also elements like line,level,... etc...
            :return: a dictionary with the structure bellow

            *Keys of returned dict*

            :group[element]:
                core value of element; also group[element_core]
            :group[element_comment]:
                comment aspect of element
            :group[element_original]:
                original aspect of element
            :group[element_str]:
                 if *include_str=True*, a string representation of elementm(with # and % if necessary);
            :group[element_kleio]:
                if *include_kleio=True*, a kleio representation in the form ``element=string``
            :group[includes]:
                list of enclosed groups
            :group[includes][subgroup]:
                list of enclosed groups of type subgroup


        If *redundant_keys=True* enclosed subgroups can also be accessed
        in the plural form, if no name conflict with existing elements::

                group['persons'] == group['includes']['persons']
                group['person']['id1'] ==
                        [p for p in group['includes']['persons']
                        if p.id='id1'][0]



        :todo: add first(), last() where includes is allowed::
                    group['first']['person'] = group['includes']['person'][0]

        """
        kd = dict()
        kd["kleio_group"] = self._name
        elements_to_include = self.elements_allowed()
        if include_builtin:
            els = elements_to_include.union(set(self._builtin_elements)) - set(["inside"])
        for e in els:
            v: KElement = getattr(self, e, None)
            if v is not None:
                if issubclass(type(v), KElement):
                    core, comment, original = v.to_tuple()
                    kd[e] = core
                    kd[e + "_core"] = core
                    kd[e + "_comment"] = comment
                    kd[e + "_original"] = original
                    if include_str:
                        kd[e + "_str"] = str(v)
                    if include_kleio:
                        kd[e + "_kleio"] = v.to_kleio()
                else:
                    kd[e] = v
        if not allow_none:
            kd = dict([(key, value) for key, value in kd.items() if value is not None])
        # we now includes subgroups
        ki = dict()
        # we now collect subgroups by name
        included = list(self.contains())
        for i in included:
            n = i.kname
            if n not in ki.keys():
                ki[n] = [
                    i.to_dict(
                        include_str=include_str,
                        include_kleio=include_kleio,
                        include_builtin=include_builtin,
                        redundant_keys=redundant_keys,
                    )
                ]
            else:
                ki[n].append(
                    i.to_dict(
                        include_str=include_str,
                        include_kleio=include_kleio,
                        include_builtin=include_builtin,
                        redundant_keys=redundant_keys,
                    )
                )
        if len(ki) > 0:
            kd["includes"] = ki
            # if there are no name conflicts and plural form
            # so g['includes']['act'] can be accessed as
            #    g['acts']
            if redundant_keys:
                for subgroup in ki.keys():
                    if subgroup + "s" not in self.elements_allowed():
                        kd[subgroup + "s"] = ki[subgroup]
                        # we include subgroup indexed by id
                        # so we can have source['act']['ac010]['person']['p01']
                        for group in ki[subgroup]:
                            gid = group.get("id", None)
                            if gid is not None and subgroup not in self.elements_allowed():  # noqa
                                if subgroup not in kd.keys():
                                    kd[subgroup] = dict()
                                kd[subgroup][gid] = group
        return kd

    def to_json(self):
        return json.dumps(
            self.to_dict(
                include_str=False,
                include_kleio=False,
                redundant_keys=False,
                include_builtin=True,
            ),
            indent=4,
            allow_nan=False,
        )

    @property
    def get(self):
        return self.to_dict()

    def to_dots(self):
        return Box(self.to_dict())

    @property
    def dots(self):
        """
        Allows easy referencing of the dictionary representation of the group.

        It is very usefull in list comprehensiona, e.g.:

          >> 'Diamantina' in [ls.value for ls in person.dots.lss
                                if ls.type == 'nome-geografico']

          >> [ls.value for ls in person.dots.lss
                        if ls.type == 'nome-geografico']

        :return:
        """
        return self.to_dots()

    def __str__(self, indent="", recurse=False):
        sname = self.kname
        s = str(sname) + "$"
        first = True
        out = []
        for e in self._position:
            v: KElement = getattr(self, e, None)
            if v is not None:
                if not first:
                    s = s + "/" + v.to_kleio(name=False)
                else:
                    s = s + v.to_kleio(name=False)
                    first = False
                out.append(e)
        more = sorted(list(set(self._guaranteed).union(set(self._also)).union(self._position).difference(out)))
        # print(more)
        if "obs" in more:  # we like obs elements at the end
            more.remove("obs")
            more.append("obs")
        for e in more:
            m: Union[KElement, str] = getattr(self, e, None)
            if m is not None and (
                type(m) is str and m > "" or (issubclass(type(m), KElement) and m.to_kleio() > "")  # noqa  # noqa
            ):
                # m contains data, lets output
                if issubclass(type(m), KElement):
                    v = m.to_kleio()
                else:
                    v = KElement(e, m).to_kleio()
                if not first:
                    s = s + f"/{v}"
                else:
                    s = s + f"{v}"
                    first = False

        if recurse:
            for g in self.contains():
                s = s + nl + g.__str__(indent + " ", recurse=recurse)
        return textwrap.indent(s, indent)

    def __getitem__(self, arg):
        if self._element_check and arg not in self.elements_allowed():
            raise ValueError("Element does not exist in group")
        return getattr(self, arg)

    def __setitem__(self, arg, value):
        if self._element_check and not self.is_allowed_as_element(arg):
            raise ValueError(f"Element not allowed: {arg}")
        el = self.pack_as_kelement(arg, value)
        setattr(self, arg, el)
        self._elementsd[arg] = el

    def pack_as_kelement(self, arg, value, element_class=None):
        """Packs value as a KElement with name arg

        :param arg: name of element
        :param value: value of element; can be a KElement, a tuple
                      (core, cooment, original) or a value
        :param element_class: class of element as string; if none arg is used to find class

        TODO: the question is what is the class of the KElement to be stored
        if value is a KElement then it already has a class
        """
        # TODO: this is wrong. if value is a KElement it should be stored as such

        if element_class is None:
            kelement = KElement.get_class_for(arg)
        else:
            kelement = KElement.get_class_for(element_class)
        if kelement is None and isinstance(value, KElement):
            if isinstance(value.element_class, KElement):
                kelement = value.element_class
            else:
                kelement = KElement.get_class_for(value.element_class)
        if kelement is None:  # if there is no KElement class we create it
            kelement = KElement.extend(arg)
            warnings.warn(
                f"Created a KElement class for {arg}. "
                f"Better to create explicit or provide "
                f" synonyms= in group creation.",
                stacklevel=2,
            )
            # we get KElement class that matches the name
            # this is how we handle localized name of elements that
            # have a builtin meaning or are referred to by standard names
            # in PomSomMapping
        if isinstance(value, KElement):
            el = kelement(value.name, core=value.core, comment=value.comment, original=value.original)
        else:
            comment = None
            original = None
            if type(value) is tuple and len(value) == 3:
                core, comment, original = value
            else:
                core = value
            el = kelement(arg, core=core, comment=comment, original=original)
        return el

    def unpack_from_kelement(self, value):
        """
        if value is a KElement return core if not return value as is.

        Useful to obtain the core value in elements that normally have
        no comment or original.

        :param value:
        :return:
        """
        if not isinstance(value, KElement):  # we did not get a KElement
            return value
        else:  # we got a KElement object
            return value.core

    def get_core(self, element, default=None):
        """get_core(element_name [, default])
        Returns the core value of an element
        """
        try:
            core = self.unpack_from_kelement(self[element])
        except ValueError:
            core = None
        except AttributeError:
            core = None
        if core is None:
            return default
        else:
            return core

    def get_id(self):
        """Return the id of the group"""
        return self.unpack_from_kelement(self.id)

    def get_element_by_name_or_class(self, element_spec, default=None):
        """
        Return the value of an element with a given name or that
        inherits from an element of the given name.

        An element matches a element_spec if its name is equal to the element_spec name
        or if it is an instance of KElement subclass with the name equal to element_spec.

        :param element_spec: name of column, or name of POMClassAttributes
        :param default: default value if not element found
        :return: KElement, or whatever in default.

        """
        el: KElement = None

        # case 1 : element_spec is a column class from a PomClassAttributes object
        #          in that case look for an element with element_class = element_spec
        elements = [el for el in self._elementsd.values() if el.element_class == element_spec]
        if len(elements) == 1:
            # a single element was found with the requested class, return it
            return elements[0]
        elif len(elements) > 1:
            # more than one element was found with the requested class.
            # this means that the mapping of this group is ambiguous regarding
            # the target table in the database, and two values can be used
            # to provide a value of a column of class element_spec.
            # We issue a warning and take the one where the element name
            # is different from the type, because it is more specialized,
            # but this is a workaround, the ambiguity should be solved at
            # the mappig level
            warnings.warn(f"Ambiguous mapping! more than one element of {self.kname}"
                          f" is mapped to column class {element_spec}"
                          f" check mapping for this group",
                          stacklevel=2)
            for el in elements:
                if el.name != element_spec:
                    # counter intuitive but normally the
                    # the best choice is the element that specialized
                    # the original one (tipo vs type, data vs date, dia vs day)
                    return el
            # if we are here more than one element matched no way to choose
            # just choose the first
            return elements[0]

        # case 2: element_spec is an element name
        for (
            name,
            el,
        ) in self._elementsd.items():  # same name as column or in ancestors # noqa: E501
            if name == element_spec:
                return el

        # case 3: element_spec is the name of one of the superclasses of an
        #         element of this group, e.g. element_spec is "dia" and this
        #         group has an element called "day" and "day" is a super class of "dia"
        for el in self._elementsd.values():  # if name in inherited names
            if element_spec in el.inherited_names():
                return el

        # case 4: element_spec and its subclasses are super class
        #         of an element of this group
        targets = KElement.get_classes_for(element_spec)
        for el in self._elementsd.values():  # check if there are alternative classes # noqa: E501
            # other classes for el
            alternatives = KElement.get_classes_for(el.name)
            # now check if there is a common path
            #   collect all the ancestors of all the classes for this el
            alt_ancestors = [(sc, sc.inherits_from()) for sc in alternatives]
            #   check if any of colspec classes are there
            for _alternative, ancestors in alt_ancestors:
                common = list(set(ancestors).intersection(set(targets)))
                if len(common) > 0:
                    return el
        return default



#!/usr/bin/env python

"""Tests for `timelink-py` package."""

import random
import pytest
from typer.testing import CliRunner
from timelink.cli import app, create_db_index, avoid_db_patterns
from tests import mhk_absent, TEST_DIR, skip_on_travis


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "Timelink and MHK manager" in result.output


@mhk_absent
def test_command_mhk_version():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(app, ["mhk", "version"])
    print(f"Test result {result.exit_code}")
    # assert result.exit_code == 0
    assert "Portainer URL" in result.output or "not found" in result.output


@mhk_absent
def test_command_mhk_status():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(app, ["mhk", "status"])
    assert result.exit_code == 0
    assert "Stopped" in result.output


# test db command
@skip_on_travis
def test_db_list():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(app, ["db", "list"])
    assert result.exit_code == 0


@skip_on_travis
def test_db_current():
    """Test the CLI."""
    runner = CliRunner()
    db_list = create_db_index(avoid_patterns=avoid_db_patterns)
    # chose a random key from dict db_list
    db_key = list(db_list.keys())[random.randint(0, len(db_list) - 1)]
    print(f"db_key: {db_key} db_list: {db_list[db_key]}")
    result = runner.invoke(app, ["db", "current", str(db_key)])
    print(f"Test result {result.exit_code}")
    print(f"Test result {result.stdout}")
    assert result.exit_code == 0


@skip_on_travis
def test_db_upgrade():
    """Test the CLI."""
    runner = CliRunner()
    db_list = create_db_index(avoid_patterns=avoid_db_patterns)
    # chose a random key from dict db_list
    db_key = list(db_list.keys())[random.randint(0, len(db_list) - 1)]
    #
    db_key = 17
    print(f"db_key: {db_key} db_list: {db_list[db_key]}")
    result = runner.invoke(app, ["db", "upgrade", str(db_key)])
    print(f"Test result {result.exit_code}")
    print(f"Test result {result.stdout}")
    assert result.exit_code == 0


@skip_on_travis
def test_db_heads():
    runner = CliRunner()
    db_list = create_db_index(avoid_patterns=avoid_db_patterns)
    # choose a random db
    db_key = list(db_list.keys())[random.randint(0, len(db_list) - 1)]
    print(f"db_key: {db_key} db_list: {db_list[db_key]}")
    heads = runner.invoke(app, ["db", "heads", str(db_key)])
    print(heads)


@skip_on_travis
def test_create_db():
    runner = CliRunner()
    result = runner.invoke(app, ["db", "create", f"sqlite:///{TEST_DIR}/db/test.db"])
    print(f"Test result {result.exit_code}")
    print(f"Test result {result.stdout}")
    assert result.exit_code == 0
    assert "created" in result.output
    # assert "not found" in result.output
    # assert "Error" in result.output
    # assert "Exception" in result.output
    # assert "Traceback" in result.output
    # assert "Usage" in result.output


def test_get_latest_version():
    from timelink import get_latest_version

    version = get_latest_version()
    print(f"Latest version: {version}")
    assert version is not None
    assert len(version) > 0



"""Test the import of Kleio files into the Timelink database.
Check tests.__init__.py for parameters

"""
# pylint: disable=import-error
import logging
import os
from pathlib import Path
from time import sleep

import pytest

from tests import TEST_DIR, get_one_translation, skip_on_travis, has_internet
from timelink.api.models.system import KleioImportedFile
from timelink.kleio.importer import import_from_xml
from timelink.api.models import base  # pylint: disable=unused-import. # noqa: F401
from timelink.api.models.base import Person, PomSomMapper
from timelink.api.database import (
    TimelinkDatabase,
    get_import_status,
)
from timelink.api.models.entity import Entity

from timelink.kleio.kleio_server import KleioServer
from timelink.kleio.schemas import KleioFile

# https://docs.pytest.org/en/latest/how-to/skipping.html
pytestmark = skip_on_travis


db_path = f"{TEST_DIR}/sqlite/"
test_set = [("sqlite", "test_import_tl"), ("postgres", "test_import_tl")]


@pytest.fixture(scope="module")
def dbsystem(request, kleio_server):
    """Create a database for testing"""
    db_type, db_name = request.param
    # only used for sqlite databases

    database = TimelinkDatabase(db_name, db_type, db_path=db_path, echo=False)
    # attach a kleio server
    database.set_kleio_server(kleio_server)  # from tests.__init__.py
    try:

        yield database
    finally:
        database.drop_db()
        database.session().close()


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_import_xml(dbsystem):
    """Test the import of a Kleio file into the Timelink database"""
    file: Path = Path(TEST_DIR, "xml_data/b1685.xml")
    with dbsystem.session() as session:

        try:
            stats = import_from_xml(file, session, options={"return_stats": True})
        except Exception as exc:
            print(exc)
            session.rolback()
            raise
        sfile: Path = stats["file"]
        assert sfile.name == file.name
        domingos_vargas = session.get(Person, "b1685.33-per6")
        assert domingos_vargas is not None, "could not get a person from file"
        kleio = domingos_vargas.to_kleio()
        assert len(kleio) > 0


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_import_linked_data_attributes(dbsystem):
    """Test the import of a translation with linked_data in attributes"""
    path = "projects/test-project/sources/reference_sources/linked_data"
    dbsystem.update_from_sources(path=path)
    with dbsystem.session() as session:
        try:
            per: Person = session.get(Person, "deh-antonio-de-abreu")
            assert per is not None, "could not get a group with linked data from file"
            k = per.to_kleio()
            print(k)
            # check for attributes with extra_info
            for a in per.attributes:
                if len(a.extra_info) > 0:
                    assert a.extra_info.keys is not None
        except Exception as exc:
            print(exc)
            session.rollback()
            raise


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_import_linked_data_geoentites(dbsystem):
    """Test the import of a translation with linked_data"""
    file: Path = Path(TEST_DIR, "xml_data/dehergne-locations-1644.xml")
    with dbsystem.session() as session:

        try:
            stats = import_from_xml(file, session, options={"return_stats": True})
        except Exception as exc:
            print(exc)
            session.rollback()
            raise
        sfile: Path = stats["file"]
        assert sfile.name == file.name
        geo1 = session.get(Entity, "deh-r1644-chekiang")
        assert geo1 is not None, "could not get a group with linked data from file"
        assert geo1.get_extra_info() is not None


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_import_atr_date(dbsystem):
    """Test the import of a Kleio file with explicit attribute dates"""
    file: Path = Path("projects/test-project/sources/reference_sources/test-attr-dates/test-attr-dates.cli")
    dbsystem.update_from_sources(file)
    with dbsystem.session() as session:
        try:
            domingos_vargas = session.get(Person, "t140337")
            assert domingos_vargas is not None, "could not get a person from file"
            kleio = domingos_vargas.to_kleio()
            assert len(kleio) > 0
        except Exception as exc:
            print(exc)
            session.rollback()
            raise


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_import_with_custom_mapping(dbsystem):
    """Test the import of a Kleio file into the Timelink database with a custom mapping"""
    file = Path("projects/test-project/sources/reference_sources/devassas/dev1692.cli")
    dbsystem.update_from_sources(file)
    with dbsystem.session() as session:
        try:
            caso = session.get(Entity, "c1692-antonio-cordeiro-alcouc")
            caso_kleio = caso.to_kleio()
            assert len(caso_kleio) > 0
            assert caso is not None, "could not get an entity from special mapping"
            caso_kleio = caso.to_kleio()
            assert len(caso_kleio) > 0
            test = session.get(Entity, "dev1692-per5")
            assert test is not None
            test_kleio = test.to_kleio()
            assert len(test_kleio) > 0
        except Exception as exc:
            print(exc)
            session.rollback()
            raise


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_import_issue48(dbsystem):
    """ Problem with importing mapping hierarchies issue #48 """
    path = "projects/test-project/sources/reference_sources/issues/issue48/"
    dbsystem.update_from_sources(path=path)
    with dbsystem.session() as session:
        try:
            carta = session.get(Entity, "deh-ca-1645-b-2-a")
            assert carta.the_type is not None, "carta did not have a type"
            pom_class: PomSomMapper = PomSomMapper.get_pom_class(carta.pom_class, session)
            assert pom_class.element_class_to_column('type', session) is not None
            crono = session.get(Entity, "issue48-sourceb-his1-79")
            assert crono.loc is not None, "crono did not have a loc"
            pom_class: PomSomMapper = PomSomMapper.get_pom_class(crono.pom_class, session)
            assert pom_class.element_class_to_column('loc', session) is not None
        except Exception as exc:
            print(exc)
            session.rollback()
            raise


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_import_with_many(dbsystem):
    """Test the import of a long Kleio file into the Timelink database"""
    file = Path("projects/test-project/sources/reference_sources/varia/auc-alunos.cli")
    dbsystem.update_from_sources(file)
    with dbsystem.session() as session:
        try:
            estudante = session.get(Entity, "140771")
            kleio = estudante.to_kleio()
            assert len(kleio) > 0
            assert estudante is not None, (
                "could not get an entity from big import"
            )  # noqa
        except Exception as exc:
            print(exc)
            session.rollback()
            raise


@pytest.mark.skipif(not has_internet(), reason="No internet connection available")
@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_import_git_hub(dbsystem):
    """Test the import of a Kleio file from github into the Timelink database"""
    file = "https://raw.githubusercontent.com/time-link/timelink-py/f76007cb7b98b39b22be8b70b3b2a62e7ae0c12f/tests/xml_data/b1685.xml"  # noqa
    with dbsystem.session() as session:
        try:
            stats = import_from_xml(file, session, options={"return_stats": True})
        except Exception as exc:
            print(exc)
            session.rollback()
            raise
        try:
            domingos_vargas = session.get(Person, "b1685.33-per6")
            assert domingos_vargas is not None, "could not get a person from file"
            kleio = domingos_vargas.to_kleio()
            assert len(kleio) > 0
            sfile: Path = stats["file"]
            assert "b1685.xml" in sfile
        except Exception as exc:
            print(exc)
            session.rollback()
            raise
        pass


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_import_from_kleio_server(dbsystem, kleio_server):
    """Test the import of a Kleio file from kleio server into the Timelink database"""
    kserver: KleioServer = kleio_server
    kleio_url = kserver.get_url()
    kleio_token = kserver.get_token()
    # wait for the server to start
    server_up = False
    counter = 0
    while not server_up and counter < 10:
        try:
            kserver.get_home_page()
            server_up = True
        except Exception as exc:
            logging.debug(exc)
            sleep(1)
            counter += 1
            pass
    # select a file randomly
    kleio_file: KleioFile = get_one_translation(kserver)
    file = kleio_file.xml_url
    with dbsystem.session() as session:
        try:
            import_from_xml(
                file,
                session,
                options={
                    "return_stats": True,
                    "kleio_token": kleio_token,
                    "kleio_url": kleio_url,
                    "mode": "TL",
                },
            )
        except Exception as exc:
            print(exc)
            session.rollback()

        try:
            # check if the file was imported
            filename_with_extension = os.path.basename(file)
            filename, extension = os.path.splitext(filename_with_extension)
            sources = session.query(KleioImportedFile).all()
            imported = False
            for source in sources:
                kfilename_with_extension = os.path.basename(source.path)
                kfile, kextention = os.path.splitext(kfilename_with_extension)
                if kfile == filename:
                    imported = True
                    break

            assert imported, "file not imported"
        except Exception as exc:
            print(exc)
            session.rollback()
            raise


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_import_status_1(dbsystem, kleio_server):
    """Test if import status is retrieved"""
    kserver: KleioServer = kleio_server
    kleio_file = get_one_translation(kserver)
    translations = [kleio_file]
    assert len(translations) > 0, "no valid translations found in Kleio Server"
    import_status = get_import_status(dbsystem, translations, match_path=True)
    assert import_status is not None



"""This module provides functions to manage a PostgreSQL server running in a Docker container.
It includes functions to check if Docker and PostgreSQL are running, retrieve PostgreSQL
container details such as user and password, construct a PostgreSQL URL, and start a PostgreSQL
server in a Docker container.

Functions:
- is_postgres_running: Check if PostgreSQL is running in Docker.
- get_postgres_container: Get the PostgreSQL container.
- get_postgres_container_pwd: Get the PostgreSQL container password.
- get_postgres_container_user: Get the PostgreSQL container user.
- get_postgres_url: Get the PostgreSQL URL for a given database name.
- start_postgres_server: Start a PostgreSQL server in Docker."""

import re
import time
import warnings
import logging
import docker
from sqlalchemy import create_engine, text
from timelink.api.database_utils import get_db_password
from timelink.kleio.kleio_server import is_docker_running

# container for postgres
postgres_container: docker.models.containers.Container = None

# SQLALCHEMY_DATABASE_URL = "sqlite:///./sql_app.db"
# SQLALCHEMY_DATABASE_URL = f"postgresql://timelink:db_password@postgresserver/db"
# SQLALCHEMY_DATABASE_URL="mysql://username:password@14.41.50.12/dbname"


def is_postgres_running():
    """Check if postgres is running in docker"""

    if not is_docker_running():
        warnings.warn("Docker is not running", stacklevel=2)
        return False

    client = docker.from_env()

    postgres_containers: list[docker.models.containers.Container] = (
        client.containers.list(filters={"ancestor": "postgres"})
    )
    return len(postgres_containers) > 0


# get the postgres container


def get_postgres_container() -> docker.models.containers.Container:
    """Get the postgres container
    Returns:
        docker.models.containers.Container: the postgres container
    """

    if not is_docker_running():
        raise RuntimeError("Docker is not running")

    client: docker.DockerClient = docker.from_env()
    postgres_containers: docker.models.container.Container = (  # pylint: disable=E1101
        client.containers.list(  # pylint: disable=E1101
            filters={"ancestor": "postgres"}
        )
    )
    return postgres_containers[0]


def get_postgres_container_pwd() -> str:
    """Get the postgres container password
    Returns:
        str: the postgres container password
    """
    if not is_docker_running():
        raise RuntimeError("Docker is not running")

    if is_postgres_running():
        container = get_postgres_container()
        pwd = [
            env
            for env in container.attrs["Config"]["Env"]
            if env.startswith("POSTGRES_PASSWORD")
        ][0].split("=")[1]
        return pwd


def get_postgres_container_user() -> str:
    """Get the postgres container user
    Returns:
        str: the postgres container user
    """
    if not is_docker_running():
        raise RuntimeError("Docker is not running")

    if is_postgres_running():
        container = get_postgres_container()
        user = [
            env
            for env in container.attrs["Config"]["Env"]
            if env.startswith("POSTGRES_USER")
        ][0].split("=")[1]
        return user


def get_postgres_url(dbname: str) -> str:
    """Get the postgres url for dbname"""
    usr = get_postgres_container_user()
    pwd = get_postgres_container_pwd()
    return f"postgresql://{usr}:{pwd}@localhost:5432/{dbname}"


def start_postgres_server(
    dbname: str | None = "timelink",
    dbuser: str | None = "timelink",
    dbpass: str | None = None,
    image: str | None = "postgres",
    version: str | None = "latest",
):
    """Starts a postgres server in docker
    Args:
        dbname (str): database name
        dbuser (str): database user
        dbpass (str): database password
        version (str | None, optional): postgres version; defaults to "latest".
    """
    if not is_docker_running():
        raise RuntimeError("Docker is not running")

    # check if postgres is already running in docker
    if is_postgres_running():
        return get_postgres_container()

    if dbname is None:
        dbname = "timelink"
    if dbuser is None:
        dbuser = "timelink"
    client = docker.from_env()
    if dbpass is None:
        dbpass = get_db_password()
    if image is None:
        image = "postgres"
    if version is None:
        version = "latest"
    psql_container = client.containers.run(
        image=f"{image}:{version}",
        detach=True,
        ports={"5432/tcp": 5432},
        environment={
            "POSTGRES_USER": dbuser,
            "POSTGRES_PASSWORD": dbpass,
            "POSTGRES_DB": dbname,
        },
    )

    timeout = 15
    stop_time = 1
    elapsed_time = 0
    # this necessary to get the status
    cont = client.containers.get(psql_container.id)
    while cont.status not in ["running"] and elapsed_time < timeout:
        time.sleep(stop_time)
        logging.debug("Waiting for postgres server to start: %s seconds", elapsed_time)
        cont = client.containers.get(psql_container.id)
        elapsed_time += stop_time
    if cont.status != "running":
        raise RuntimeError("Postgres server did not start")

    while True:
        # Execute the 'pg_isready' command in the container
        exit_code, _output = psql_container.exec_run("pg_isready")

        # If the 'pg_isready' command succeeded, break the loop
        if exit_code == 0:
            logging.info("Postgres server is ready")
            break

        # If the 'pg_isready' command failed, wait for 1 seconds and try again
        time.sleep(1)

    return psql_container


def get_postgres_dbnames():
    """Get the database names from a postgres server
    Returns:
        list[str]: list of database names

    ..code-block:: sql

        SELECT datname
            FROM pg_database
        WHERE NOT datistemplate
                AND datallowconn
                AND datname <> 'postgres';
    """

    if not is_docker_running():
        raise RuntimeError("Docker is not running")

    container = start_postgres_server()
    if container is not None:
        engine = create_engine(
            f"postgresql://{get_postgres_container_user()}:{get_postgres_container_pwd()}@localhost:5432/postgres"
        )
        with engine.connect() as conn:
            dbnames = conn.execute(
                text(
                    "SELECT datname FROM pg_database "
                    "WHERE NOT datistemplate AND datallowconn AND datname <> 'postgres';"
                )
            )
            result = [dbname[0] for dbname in dbnames]
        return result
    return []


def is_valid_postgres_db_name(db_name):
    """Check if the database name is valid"""
    # Check if the name is less than 64 characters long
    if len(db_name) >= 64:
        return False

    # Check if the name starts with a letter or underscore
    if not re.match(r"^[a-zA-Z_]", db_name):
        return False

    # Check if the name contains only letters, digits, and underscores
    if not re.match(r"^\w+$", db_name):
        return False

    # If all checks pass, the name is valid
    return True





# tests/test_utilities.py
# pylint: disable=C0116

from timelink.kleio.utilities import (
    get_extra_info_from_obs,
    quote_long_text,
    kleio_escape,
    render_with_extra_info
)


def test_quote_long_text_with_none():
    assert quote_long_text(None) is None


def test_quote_long_text_with_short_text():
    assert quote_long_text("short text") == "short text"


def test_quote_long_text_with_long_text():
    long_text = "This is a very long text that should be quoted because it exceeds the length limit."
    s = quote_long_text(long_text, width=64)
    assert len(s.splitlines()) > 1


def test_quote_long_text_with_text_containing_quotes():
    text_with_quotes = 'This text "contains" quotes.'
    s = quote_long_text(text_with_quotes)
    assert s == f'"""{text_with_quotes}"""'


def test_quote_long_text_with_text_containing_newlines():
    text_with_newlines = "This text\ncontains\nnewlines."
    s = quote_long_text(text_with_newlines, initial_indent="", indent="", width=80)
    assert s == f'"""{text_with_newlines}\n"""'


def test_quote_long_text_with_text_containing_special_chars():
    text_with_special_chars = "This text contains special chars like /, ;, =, #, $, % but no quotes nor newlines"
    s = quote_long_text(text_with_special_chars, initial_indent="", indent="", width=120)
    assert s == f'"{text_with_special_chars}"'


def test_get_extra_info_with_none():
    result = get_extra_info_from_obs(None)
    assert result == ("", {})


def test_get_extra_info_with_empty_string():
    result = get_extra_info_from_obs("")
    assert result == ("", {})


def test_get_extra_info_with_valid_string():
    extra_info = 'saiu a primeira vez extra_info: {"value": {"comment": "@wikidata:Q45412"}}'
    expected_result = ("saiu a primeira vez", {"value": {"comment": "@wikidata:Q45412"}})
    result = get_extra_info_from_obs(extra_info)
    assert result == expected_result


def test_kleio_escape_with_none():
    assert kleio_escape(None) is None


def test_kleio_escape_with_no_special_chars():
    assert kleio_escape("simpletext") == "simpletext"
    assert kleio_escape("@wikidata:Q45412") == "@wikidata:Q45412"


def test_kleio_escape_with_special_chars():
    assert kleio_escape("text/with/special;chars") == '"text/with/special;chars"'
    assert kleio_escape("text=with=special#chars") == '"text=with=special#chars"'
    assert kleio_escape("text$with%special\nchars") == '"text$with%special\nchars"'


def test_kleio_escape_with_quotes():
    quoted = kleio_escape('text"with"quotes')
    assert quoted == '"""text"with"quotes"""'


def test_kleio_escape_with_mixed_chars():
    assert kleio_escape("text/with=special;chars#and$more%chars\n") == '"text/with=special;chars#and$more%chars\n"'


def test_render_with_extra_info():
    element_name = "value"
    element_value = 'saiu a primeira vez extra_info: {"value": {"comment": "@wikidata:Q45412"}}'
    core, extra_info = get_extra_info_from_obs(element_value)
    expected_result = f'{core}#@wikidata:Q45412'
    assert render_with_extra_info(element_name, core, extra_info) == expected_result



"""Test the import of Kleio files into the Timelink database.
Check tests.__init__.py for parameters

"""

# pylint: disable=import-error
from pathlib import Path
import random
from typing import List

import pytest
from sqlalchemy import select
from sqlalchemy_utils import drop_database

from tests import TEST_DIR, skip_on_travis
from timelink.api.models.relation import Relation
from timelink.api.models.system import KleioImportedFile
from timelink.kleio.kleio_server import KleioFile, KleioServer
from timelink.kleio.importer import import_from_xml
from timelink.api.models import base  # pylint: disable=unused-import. # noqa: F401
from timelink.api.models.base import Person
from timelink.api.models.rentity import Link, REntity, LinkStatus as STATUS
from timelink.api.database import (
    TimelinkDatabase,
    start_postgres_server,
    get_postgres_container_user,
    get_postgres_container_pwd,
)

# https://docs.pytest.org/en/latest/how-to/skipping.html
pytestmark = skip_on_travis

db_path = Path(TEST_DIR, "sqlite")
RENTITY_DB = "rentities"
# rentity_db = ":memory:"

# set a list of files to be imported before the tests begin
# relative to KLEIO_HOME
TEST_FILES_DIR = "projects/test-project/sources/reference_sources/rentities"


@pytest.fixture(scope="module")
def dbsystem(request, kleio_server):
    """Create a database for testing"""
    db_type, db_name, db_url, db_user, db_pwd = request.param
    if db_type == "postgres":
        start_postgres_server()
        db_user = get_postgres_container_user()
        db_pwd = get_postgres_container_pwd()

    database = TimelinkDatabase(
        db_name=db_name,
        db_type=db_type,
        db_path=db_path,
        db_url=db_url,
        db_user=db_user,
        db_pwd=db_pwd,
    )

    database.set_kleio_server(kleio_server)

    try:
        yield database
    finally:
        database.drop_db()
        database.session().close()
        if ":memory:" not in database.db_url:
            drop_database(database.db_url)


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", RENTITY_DB, None, None, None),
        # change to pytest.param("postgres", "rentities", None, None, None, marks=skip_on_travis)
        # to skip the test on travis see https://doc.pytest.org/en/latest/how-to/skipping.html#skip-xfail-with-parametrize
        ("postgres", "rentities", None, None, None),
    ],
    indirect=True,
)
def test_link_two_occ(dbsystem):
    """Test the linking of two occurrences

    Will test all the possible combinations:
        two unbound occurrences
        first one bound second unbound
        first one unboud second bound
        both bound to same real entity
        both bound to separate real entities

    Also tests deletion of real entities
    and determining the real entity of an occurrence
    """
    with dbsystem.session() as session:

        session.expire_on_commit = False

        file1 = Path(TEST_FILES_DIR, "sameas-tests.cli")
        dbsystem.update_from_sources(file1, force=True)

        ricci = session.get(Person, "deh-matteo-ricci")
        assert ricci is not None, "could not get a person from file"
        ricci.to_kleio()

        test_rid = "rp-matteo"

        REntity.delete(test_rid, session=session)

        # Fetch occurrences
        # everyboby with a name like Matteo Ricci
        occurrences = [id for (id,) in session.query(Person.id).filter(Person.name.like("Mat%Ricci")).all()]

        # Shuffle the occurrences to randomize the order
        random.shuffle(occurrences)

        # erase any previous real entities
        # select distinct real_id from links where entity_id in (occurrences)
        stmt = select(Link.rid).where(Link.entity.in_(occurrences)).distinct()

        rids = session.execute(stmt).scalars().all()

        for real_id in rids:
            REntity.delete(real_id, session=session)
            link = session.query(Link).where(Link.rid == real_id).all()
            print(link)

        # test all the combinations
        occ1 = occurrences[0]
        occ2 = occurrences[1]
        # two unbound occurrences
        ri1 = REntity.same_as(occ1, occ2, real_id=test_rid, status=STATUS.MANUAL, session=session)
        assert ri1 is not None, "Real entity not returned"
        assert ri1.id == test_rid, "real_id not preserved"

        occ3 = occurrences[2]
        # first one bound second unbound
        ri2 = REntity.same_as(occ2, occ3, status=STATUS.AUTOMATIC, session=session)
        assert ri2.id == ri1.id
        # first one unboud second bound
        occ4 = occurrences[3]
        r3 = REntity.same_as(occ4, occ3, status=STATUS.AUTOMATIC, session=session)
        assert r3.id == ri1.id

        # both bound to same real entity
        r4 = REntity.same_as(occ4, occ3, status=STATUS.AUTOMATIC, session=session)
        assert r4.id == ri1.id

        # both bound to separate real entities
        occ5 = occurrences[4]
        occ6 = occurrences[5]
        r5 = REntity.same_as(occ5, occ6, status=STATUS.AUTOMATIC, session=session)
        assert r5.id != ri1.id
        # now merge the two real entities
        r6 = REntity.same_as(occ5, occ1, status=STATUS.AUTOMATIC, session=session)
        assert r6.id == ri1.id

        r5 = session.get(REntity, r5.id)
        assert r5 is None

        # check that the real entity has the correct number of occurrences
        rel_entity = session.get(REntity, r6.id)
        occs = [link.entity for link in rel_entity.links]
        assert len(occs) == 6
        assert occ1 in occs
        assert occ2 in occs
        assert occ3 in occs
        assert occ4 in occs
        assert occ5 in occs
        assert occ6 in occs

        # check that subsequent calls to same_as return the same real entity
        r7 = REntity.same_as(occ5, occ6, status=STATUS.AUTOMATIC, session=session)
        assert r7.id == r6.id

        # check that cannot set real_id if occurrences are already bound
        with pytest.raises(ValueError):
            REntity.same_as(occ5, occ6, real_id="xpto", status=STATUS.AUTOMATIC, session=session)

        real_ricci = session.get(REntity, ri1.id)
        assert len(real_ricci.to_kleio()) > 0
        real_ricci2 = REntity.get(ri1.id, session=session)
        assert real_ricci.id == real_ricci2.id


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", RENTITY_DB, None, None, None),
        # change to pytest.param("postgres", "rentities", None, None, None, marks=skip_on_travis)
        # to skip the test on travis see https://doc.pytest.org/en/latest/how-to/skipping.html#skip-xfail-with-parametrize
        ("postgres", "rentities", None, None, None),
    ],
    indirect=True,
)
def test_make_real(dbsystem):
    """Test the creation of a real entity"""
    dbsystem.update_from_sources(Path(TEST_FILES_DIR, "sameas-tests.cli"))
    with dbsystem.session() as session:
        bento_de_gois = session.get(Person, "deh-bento-de-gois")
        assert bento_de_gois is not None, "could not get a person from file"

        rid = REntity.make_real(bento_de_gois.id, status=STATUS.AUTOMATIC, session=session)
        assert rid is not None, "real_id not returned"
        real_bento = session.get(REntity, rid.id)
        assert len(real_bento.to_kleio()) > 0


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", "xsameas", None, None, None),
        # change to pytest.param("postgres", "rentities", None, None, None, marks=skip_on_travis)
        # to skip the test on travis see https://doc.pytest.org/en/latest/how-to/skipping.html#skip-xfail-with-parametrize
        ("postgres", "xsameas", None, None, None),
    ],
    indirect=True,
)
def test_import_xsameas(dbsystem: TimelinkDatabase):
    # Test that the sequence of import files does not affect the result
    file1 = f"{TEST_FILES_DIR}/sameas-tests.cli"
    file2 = f"{TEST_FILES_DIR}/xsameas-tests.cli"

    with dbsystem.session() as session:

        dbsystem.update_from_sources(file1)
        dbsystem.update_from_sources(file2)

        # check the number of KleioFiles imported
        stmt = select(KleioImportedFile)

        kfiles = session.execute(stmt).scalars().all()
        print(len(kfiles))

        # count the number of links
        stmt = select(Link)
        links = session.execute(stmt).scalars().all()
        linked_entities = sorted([link.entity for link in links])
        linked_rentities = sorted([link.rid for link in links])

        nlinks = len(links)

        # the problem here is that 'bio-michele-ruggieri-his4-19-per1-21'
        # occurs twice in the link list
        # ('rp744201',
        # 'bio-michele-ruggieri-his4-19-per1-21',
        # 'xsameas-tests')
        # this a bug in the sameas processing
        # count the number of relations with None in destination
        stmt = select(Relation).where(Relation.destination is None)
        relations = session.execute(stmt).scalars().all()
        nrelations = len(relations)

        # We now reimport file1 and check if the number of links and relations is the same
        importfile1: List[KleioFile] = dbsystem.kserver.get_translations(file1)
        xml_file = importfile1[0].xml_url
        import_from_xml(
            xml_file,
            session,
            options={
                "return_stats": True,
                "kleio_token": dbsystem.kserver.get_token(),
                "kleio_url": dbsystem.kserver.get_url(),
                "mode": "TL",
            },
        )

        stmt = select(Link)
        links2 = session.execute(stmt).scalars().all()
        linked_entities_2 = sorted([link.entity for link in links2])
        linked_rentities_2 = sorted([link.rid for link in links2])

        diff = set(linked_entities) - set(linked_entities_2)
        re_diff = set(linked_rentities) - set(linked_rentities_2)
        print(diff, re_diff)
        # this fails because the redundant link above was
        # removed in this phase.
        assert len(links) == nlinks, "wrong number of links"
        assert len(relations) == nrelations, "wrong number of relations"
        assert len(diff) == 0, "wrong number of occurrences linked"
        assert len(re_diff) == 0, "wrong number of rentities linked"

        stmt = select(Relation).where(Relation.destination == None)  # noqa
        relations2 = session.execute(stmt).scalars().all()
        assert len(relations2) == nrelations, "some relations have no destination after reimport"


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", RENTITY_DB, None, None, None),
        # change to pytest.param("postgres", "rentities", None, None, None, marks=skip_on_travis)
        # to skip the test on travis see https://doc.pytest.org/en/latest/how-to/skipping.html#skip-xfail-with-parametrize
        ("postgres", "rentities", None, None, None),
    ],
    indirect=True,
)
def test_import_aregister(dbsystem: TimelinkDatabase):
    """Test the import a identifications file"""
    # first fill in the database with the sameas file
    file1 = Path(TEST_FILES_DIR, "sameas-tests.cli")
    dbsystem.update_from_sources(file1, force=True)
    # now import the identifications file
    file = Path(TEST_FILES_DIR, "aregister-tests.cli")
    dbsystem.update_from_sources(file, force=True)
    kserver: KleioServer = dbsystem.kserver
    kfiles = kserver.get_translations(path=file, status='V')
    assert len(kfiles) == 1, "wrong number of files"
    kfile = kfiles[0]
    stats = dbsystem.import_from_xml(kfile)
    # Import again to test the update mechanis
    sfile = stats["file"]
    assert "aregister" in sfile  # ensure import was done
    with dbsystem.session() as session:
        real_person = session.get(REntity, "rp-66")
        assert real_person is not None, "could not get a real person from identifications import"  # noqa
        assert len(real_person.get_occurrences()) == 10, "wrong number of occurrences"
        kleio = real_person.to_kleio()
        assert len(kleio) > 0
        stats = dbsystem.import_from_xml(kfile)
        sfile = stats["file"]
        assert "aregister" in sfile
        real_person = session.get(REntity, "rp-66")
        assert real_person is not None, "could not get a real person from identifications import"
        assert len(real_person.get_occurrences()) == 10, "wrong number of occurrences"


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", RENTITY_DB, None, None, None),
        # change to pytest.param("postgres", "rentities", None, None, None, marks=skip_on_travis)
        # to skip the test on travis see https://doc.pytest.org/en/latest/how-to/skipping.html#skip-xfail-with-parametrize
        ("postgres", "rentities", None, None, None),
    ],
    indirect=True,
)
def test_random_id(dbsystem):
    """Test the generation of a random id"""
    with dbsystem.session() as session:
        id1 = REntity.generate_id(session=session)
        id2 = REntity.generate_id(session=session)
        assert id1 != id2



# blake disable unused import TODO: not sure the purpose of this file. see __init__.py
from .base_class import Base  # noqa pylint: disable=unused-import
from .entity import Entity  # noqa pylint: disable=unused-import
from .attribute import Attribute  # noqa pylint: disable=unused-import
from .relation import Relation  # noqa pylint: disable=unused-import
from .person import Person  # noqa pylint: disable=unused-import
from .object import Object  # noqa pylint: disable=unused-import
from .source import Source  # noqa pylint: disable=unused-import
from .act import Act  # noqa pylint: disable=unused-import
from .rentity import REntity  # noqa pylint: disable=unused-import
from .rentity import LinkStatus  # noqa pylint: disable=unused-import
from .pom_som_mapper import PomSomMapper  # noqa pylint: disable=unused-import
from .pom_som_mapper import PomClassAttributes  # noqa pylint: disable=unused-import



from enum import Enum as PyEnum
import random
import math
from typing import Optional
from itertools import chain

from sqlalchemy import Enum, Integer, String, ForeignKey, update
from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.orm import object_session

from .base_class import Base
from .entity import Entity
from .aregister import ARegister
from .source import Source


class OccurrenceMissingError(ValueError):
    """Error raised when an occurrence is missing in the database
    and an attempt is made to link it to a real entity"""

    pass


class RealEntityMissingError(ValueError):
    """Error raised when a Real Entity required for an opperation is missing"""

    pass


class OccurenceTypeError(ValueError):
    """Error raised when an attempot is made to link two entities of different types"""

    pass


class RealEntityIdExists(ValueError):
    """Error raised when an attempt is made to create a real entity with an existing id"""

    pass


class RealEntityIdChangeError(ValueError):
    """Error raised when an attempt is made to change the id of a real entity"""

    pass


class LinkStatus(PyEnum):
    """
    Status of a real entity.

    The following statuses are possible:

    Unlinked: unlinked entity, assumed as a real entity.
    Automatic: linked automatically by enetity resolution algorithms.
    Source: linked by the user in the source with same_as or x_same_as.
    Manual: linked manually by the user in the database after import.
    Valid: validated by the user (whatever the origin).
    N: legacy status, linked in MHK (considered valid). For importing data from MHK.

    Note that in SQLAlchemy what is persisted is the name of the enum
    "UNLINKED", "AUTOMATIC", "SOURCE", "MANUAL", "VALID", "MHK" etc...

    Note that changes to this list will require changes to the database schema in postgresql
    directly in the database.

    See: https://stackoverflow.com/questions/43519096/
                after-updating-enum-with-a-new-value-that-value-cannot-be-inserted-with-psycopg
    """

    # U=linked (single occurrence), A=Automatic, S=Source, M=Manual linked, V=Valid
    UNLINKED = "0"  # unlinked entity
    AUTOMATIC = "1"  # linked automatically
    SOURCE = "2"  # linked in the source with same_as or x_same_as
    MANUAL = "3"  # Linked manually by the user in the database after import
    MHK = "N"  # legacy status, linked in MHK (considered valid)
    N = "MHK"  # legacy status, linked in MHK (considered valid)
    INVALID = "I"  # During import of a source, links are invalidated

    def __str__(self):
        return self.name


class REntity(Entity):
    """
    Represents Real Entities in the database.

    Real entities are the result of entity resolution.
    They aggregate the information of several entities
    in the database that are believed to represent the same real-world entity.

    The representation supports reversing the entity resolution
    process. A mapping between the real entity and the entities
    that compose is maintained in the `links` table.

    We use the term "REntity" to avoid confusion with the term "entity",
    and often use to unlinked entities as "occurrences".

    Note that it is possible for an occurrence to be linked
    to more than one real entity. This can happen when the
    different users have different opinions about the identity
    of the entities.

    Fields:
        id: id of the real entity
        user: user that identified this real entity
        description: description of the real entity
        status: status of the real entity
        obs: observations about the real entity
        links: list of links to the entities that compose the real entity


    """

    __tablename__ = "rentities"
    __mapper_args__ = {"polymorphic_identity": "rentity"}

    id: Mapped[str] = mapped_column(
        String(64), ForeignKey(Entity.id, ondelete="CASCADE"), primary_key=True
    )
    user: Mapped[str] = mapped_column(
        String(64)
    )  # user that identified this real entity
    description: Mapped[str] = mapped_column(
        String(4096)
    )  # description of the real entity
    status: Mapped[LinkStatus] = mapped_column(
        Enum(LinkStatus), nullable=False
    )  # status of the real entity
    obs: Mapped[Optional[str]] = mapped_column(
        String
    )  # observations about the real entity

    links: Mapped[list["Link"]] = relationship(
        "Link",
        back_populates="rentity_rel",
        passive_deletes=True,
    )  # One-to-many relationship

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"REntity(id={sr}, "
            f"user={self.user}, "
            f"description={self.description}, "
            f"status={self.status}, "
            f"obs={self.obs}"
            f")"
        )

    # introspection
    def get_user(self):
        """Get the user that identified this real entity"""
        return self.user

    def set_user(self, user):
        """Set the user that identified this real entity"""
        self.user = user

    def get_description(self):
        """Get the description of the real entity"""
        return self.description

    def set_description(self, description):
        """Set the description of the real entity"""
        self.description = description

    def get_status(self):
        """Get the status of the real entity"""
        return self.status

    def set_status(self, status):
        """Set the status of the real entity"""
        self.status = status

    def get_obs(self):
        """Get the observations about the real entity"""
        return self.obs

    def set_obs(self, obs):
        """Set the observations about the real entity"""
        self.obs = obs

    def get_occurrences(self):
        """Get the occurrences of the real entity"""
        return [link.entity for link in self.links]

    @property
    def contains(self):
        # get the sqlalchemy session of self
        session = object_session(self)
        # collect all the entities that are linked to this real entity and return the ones they contain
        return list(
            chain.from_iterable(
                session.get(Entity, occ.entity).contains for occ in self.links
            )
        )

    @property
    def attributes(self):
        # get the sqlalchemy session of self
        session = object_session(self)
        # collect all the entities that are linked to this real entity and return the ones they contain
        return list(
            chain.from_iterable(
                session.get(Entity, occ.entity).attributes for occ in self.links
            )
        )

    @property
    def rels_in(self):
        """Relations having this real entity as destination"""
        # get the sqlalchemy session of self
        session = object_session(self)
        return list(
            chain.from_iterable(
                session.get(Entity, occ.entity).rels_in for occ in self.links
            )
        )

    @property
    def rels_out(self):
        """Relations having this real entity as source"""
        # get the sqlalchemy session of self
        session = object_session(self)
        return list(
            chain.from_iterable(
                session.get(Entity, occ.entity).rels_out for occ in self.links
            )
        )

    @classmethod
    def get_rentity_brief(cls, rentity_id: str, session=None, db=None):
        """
        Fetch a real entity from the database
        """
        if session is not None:
            r: REntity = session.get(REntity, rentity_id)
        elif db is not None:
            with db.session() as session:
                r: REntity = session.get(REntity, rentity_id)
        else:
            raise ValueError("Error, session or db needed")
        return r

    @classmethod
    def get_rentity_full(cls, rentity_id: str, session=None, db=None):
        """
        Fetch a real entity from the database with all the
        attributes, relationships and contained objects of
        the occurrences of the real entity.

        I think this corresponds to an eager fetch in SQLAlchemy
        """

        if session is not None:
            # Fetch eargely the occurrences

            r: REntity = session.get(REntity, rentity_id)
        elif db is not None:
            with db.session() as session:
                r: REntity = session.get(REntity, rentity_id)
        else:
            raise ValueError("Error, session or db needed")

        return r

    @classmethod
    def get(cls, rentity_id: str, session=None):
        """Get a real entity from the database"""

        if session is None:
            raise ValueError("Error, session needed")

        r: REntity = session.get(REntity, rentity_id)
        return r

    @classmethod
    def already_real(cls, occurrence: str, user="user", session=None):
        """Check if an occurrence is already linked to a real entity
        for a given user"""
        if session is None:
            raise ValueError("Error, session needed")

        real = (
            session.query(Link)
            .filter(Link.entity == occurrence, Link.user == user)
            .all()
        )
        return len(real) > 0

    @classmethod
    def get_real_entity(cls, occurrence: str, user="user", session=None):
        """Get the real entity of an occurrence
        TODO: this should be an instance method of Entity
        """
        if session is None:
            raise ValueError("Error, session needed")

        links = (
            session.query(Link)
            .filter(Link.entity == occurrence, Link.user == user)
            .all()
        )
        if len(links) == 0:
            return None
        return links[0].rid

    @classmethod
    def delete(cls, rentity_id: str, silent=True, session=None):
        """
        Delete a real entity from the database

        If sucessfull returns the id of the deleted real entity
        if the rentity does not exist and silent=True returns None otherwise raises an exception
        """

        if session is None:
            raise ValueError("Error, session needed")

        # session.commit()
        # session.begin()
        # delete the real entity
        try:
            re = session.get(REntity, rentity_id)
            if re is None:
                if not silent:
                    raise ValueError(f"Error, {rentity_id} does not exist")
                return None
            else:
                for link in re.links:
                    session.delete(link)
                session.delete(re)
        except Exception as e:
            session.rollback()
            raise e
        session.commit()
        return rentity_id

    @classmethod
    def same_as(
        cls,
        id1: str,
        id2: str,
        user="user",
        status=None,
        real_id=None,
        real_id_prefix=None,
        description=None,
        rule=None,
        source=None,
        session=None,
    ) -> str:
        """Returns a real entity id linking id1 and id2.

        Args:
            id1: id of the first entity
            id2: id of the second entity
            user: user that linked the entities
            status: status of the link (default M)
            real_id: id of the real entity (if none a random id is generated)
            real_id_prefix: prefix of the generated real entity id
            description: description of the real entity (defaults to description or name of first entity)
            rule: rule used to link the entities
            source: if of the source with the link (same_as or x_same_as)
            session: database session

        if id1 and id2 are both unliked occurrences then a new real entity is created
            associated with user.
            if real_id is not given a random id is generated.

        if id1 is a real entity and id2 is not, then id2 is added to id1 and id1 returned
        if id2 is a real entity and id1 is not, swap and do as above.
        if both id1 and id2 are real entities, merge them and keep the id of the real
            entity with higher status V->M->S->A->U

        in all the cases linking two id2 of different types or inheriting from a
        common type not equal to Entity is an error."""
        if session is None:
            raise ValueError("Error, session needed")

        if "bio-michele-ruggieri-his4-19-per1-21" in [id1, id2]:
            print(f"DEBUG same_as({id1}, {id2})")

        # session.commit()
        # session.begin()
        if id1 == id2:
            return cls.make_real(
                id1,
                user=user,
                status=status,
                real_id=real_id,
                real_id_prefix=real_id_prefix,
                description=description,
                rule=rule,
                source=source,
                session=session,
            )

        # check if Entity id1 and id2 exists and are of the same type
        eid1 = session.get(Entity, id1)
        eid2 = session.get(Entity, id2)
        if eid1 is None or eid2 is None:
            raise OccurrenceMissingError(
                f"Error, {id1} and {id2} must exist in the database"
            )

        if eid1.pom_class != eid2.pom_class:
            raise OccurenceTypeError(
                f"Error, {id1} and {id2} must be of the same type"
            )

        # check if id1 and id2 are already linked
        # Query the links table for entity=id1 and user=user and return the id, rid, and status

        result = (
            session.query(Link.rid, Link.id, Link.status)
            .filter(Link.entity == id1, Link.user == user)
            .first()
        )
        if result is None:
            result = (None, None, None)
        r1_id, l1_id, l1_status = result

        result = (
            session.query(Link.rid, Link.id, Link.status)
            .filter(Link.entity == id2, Link.user == user)
            .first()
        )
        if result is None:
            result = (None, None, None)
        r2_id, l2_id, l2_status = result

        # CASE 1: occurrences are already linked to the same real entity
        if (
            r1_id is not None
            and r2_id is not None  # noqa: W503
            and r1_id == r2_id  # noqa: W503
        ):
            # Check if it is possible to change the real_id
            if real_id is not None and r1_id != real_id:
                raise ValueError(
                    f"Error, {id1} and {id2} are already linked to {r1_id}"
                )
            return session.get(REntity, r1_id)

        if rule is None:
            rule = f"same_as('{id1}', '{id2}')"

        if status is None:
            status = LinkStatus.MANUAL

        # CASE 2: occurrences are not linked
        # New real entity is created or reused
        if r1_id is None and r2_id is None:
            # both are unlinked occurrences

            if description is None:
                desc = "<No description>"
                if hasattr(eid1, "description"):
                    desc = eid1.description
                elif hasattr(eid1, "name"):
                    desc = eid1.name  # TODO check any column of class "Name"
            else:
                desc = description

            # check if the two entities were previously
            # to a real entity, if so reuse the real entitiy information
            br1 = cls.recover_rentity(id1, user=user, session=session)
            br2 = cls.recover_rentity(id2, user=user, session=session)

            if br1 is not None and br2 is not None and br1 == br2:
                # we reuse the real entity
                r = session.get(REntity, br1)
            else:
                if real_id is None:
                    if real_id_prefix is None:
                        ridp = (
                            "r" + eid1.pom_class[0]
                        )  # we take first letter of the class
                    else:
                        ridp = real_id_prefix
                    real_id = cls.generate_id(session=session)
                    real_id = f"{ridp}{real_id}"
                else:
                    # check if real_id exists
                    if session.get(REntity, real_id) is not None:
                        raise RealEntityIdExists(f"Error, {real_id} already exists")

                r = REntity(id=real_id, user=user, description=desc, status=status)
                session.add(r)

            l1 = Link(
                rid=real_id,
                entity=id1,
                user=user,
                status=status,
                rule=rule,
                source=source,
            )
            l2 = Link(
                rid=real_id,
                entity=id2,
                user=user,
                status=status,
                rule=rule,
                source=source,
            )
            r.links.append(l1)
            r.links.append(l2)
            session.flush()
            session.commit()

        # CASE 3: one of the occurrences is already linked
        elif r1_id is not None and r2_id is None:  # noqa: W503
            # r1 is a real entity and r2 is an occurrence
            if real_id is not None:
                if r1_id.id != real_id:
                    raise RealEntityIdChangeError(
                        f"Error, {id1} is already linked to {r1_id}, cannot change to {real_id}. "
                        f"Use make_real('{id2}', real_id='{real_id}')"
                    )
            # validate the first linlk
            l1 = session.get(Link, l1_id)
            l1.status = status
            session.merge(l1)
            # add the second link
            r = session.get(REntity, r1_id)

            link = Link(
                rid=r1_id,
                entity=id2,
                user=user,
                status=status,
                rule=rule,
                source=source,
            )
            r.links.append(link)
            session.merge(r)
            session.commit()

        # CASE 4: one of the occurrences is already linked (right)
        elif r1_id is None and r2_id is not None:  # noqa: W503
            # rid2 is a real entity and id1 is an occurrence
            if real_id is not None:
                if r2_id != real_id:
                    raise RealEntityIdChangeError(
                        f"Error, {id2} is already linked to {r2_id}, cannot change to {real_id}. "
                        f"Use make_real('{id1}', real_id='{real_id}')"
                    )
            # validate first link
            r = session.get(REntity, r2_id)
            l2 = session.get(Link, l2_id)
            l2.status = status
            session.merge(l2)
            link = Link(
                rid=r2_id,
                entity=id1,
                user=user,
                status=status,
                rule=rule,
                source=source,
            )
            r.links.append(link)

        # both are real entities
        elif (r1_id is not None and r2_id is not None):  # noqa: W503
            real1 = session.get(REntity, r1_id)
            real2 = session.get(REntity, r2_id)

            if real1.status.value > real2.status.value or (  # noqa: W504
                real1.status.value == real2.status.value and real1.id < real2.id
            ):
                # r1 is more important
                # update links with rid=r2 to rid=r1
                # update with SQLALchemy update statement
                # update links set rid = r1 where rid = r2
                stmt = (
                    update(Link)
                    .where(Link.rid == r2_id)
                    .values(rid=r1_id, rule=rule, source=source)
                )
                session.execute(stmt)
                session.commit()
                session.expunge(real1)
                session.expunge(real2)
                REntity.delete(r2_id, silent=False, session=session)
                r = session.get(REntity, r1_id)
            else:
                # r2 is more important
                # update links with rid=r1 to rid=r2
                # update with SQLALchemy update statement
                # update links set rid = r2 where rid = r1
                stmt = (
                    update(Link)
                    .where(Link.rid == r1_id)
                    .values(rid=r2_id, rule=rule, source=source)
                )
                session.execute(stmt)
                session.commit()
                session.expunge(real1)
                session.expunge(real2)
                REntity.delete(r1_id, silent=False, session=session)
                r = session.get(REntity, r2_id)
        else:
            raise ValueError(
                f"Error, could not determine same as strategy for {id1} and {id2} "
                f" are already linked to {r1_id} and {r2_id} with link status {l1_status} and {l2_status}"
            )
        session.commit()
        return r

    @classmethod
    def generate_id(cls, length=6, session=None):
        """Generate a random n digit id for a real entity"""
        if session is None:
            raise ValueError("Error, session needed")

        # storing strings in a list
        digits = [i for i in range(0, 10)]

        # initializing a string
        random_str = ""

        circuit_breaker = 1000000
        while True:
            circuit_breaker -= 1
            if circuit_breaker < 0:
                raise ValueError(
                    f"Error, could not generate a unique id with {length} digits"
                )
            random_str = ""
            for _i in range(length):
                # generating a random index
                index = math.floor(random.random() * 10)
                random_str += str(digits[index])
            # check if the id already exists
            if session.get(Entity, random_str) is None:
                break  # if not return
        return random_str

    @classmethod
    def recover_rentity(cls, occ: str, user="user", session=None):
        """Recover a real entity from an occurrence

        This method is used to recover a real entity an occurrence
        that were previously linked to the real entity. It is used to avoid
        the creation of a new real entity when occurrences were
        temporarlily unlinked and then linked again, which is a common
        operation when reimporting data from a source.

        This method uses the blinks table, which are backed up links
        saved before a source is reimported.

        If the occurence was previously linked to a real entity
        and currently the real entity has no links to entities
        then the id of the real entity is returned

        Args:
            id1: id of the first occurrence
            user: user that linked the occurrences
            session: database session

        Returns:
            The real entity id if the occurrence was linked to the real entity in the past
            and the real entity is not linked to any other entity
        """
        if session is None:
            raise ValueError("Error, session needed")

        # session.commit()
        # session.begin()
        # check if Entity id1 and id2 exists and are of the same type
        eid1 = session.get(Entity, occ)
        if eid1 is None:
            raise OccurrenceMissingError(f"Error, {occ} must exist in the database")

        # check if id1 is already linked
        # Query the links table for entity=id1 and user=user and return the id,
        # rid, and status

        r1_id = (
            session.query(BLink.rid)
            .filter(BLink.entity == occ, BLink.user == user)
            .scalar()
        )

        if r1_id is not None:
            rentity = session.get(REntity, r1_id)
            if rentity is not None:
                result = r1_id
            else:
                result = None
        else:
            result = None
        return result

    @classmethod
    def make_real(
        cls,
        id1: str,
        user="user",
        status=LinkStatus.MANUAL,
        real_id=None,
        real_id_prefix=None,
        description=None,
        rule=None,
        source=None,
        session=None,
    ):
        """Make an entity a real entity

        This creates a real entity from a single occurence

        Args:
            id1: id of the occurrence
            user: user that linked the occurrence
            status: status of the link
            real_id: id of the real entity
            real_id_prefix: prefix of the generated real entity id
            description: description of the real entity
            rule: rule used to link the occurrence
            source: id of the source of the link (normally an identifications source)
            session: database session

        Returns:
            The real entity created
        """
        if session is None:
            raise ValueError("Error, session needed")

        if rule is None:
            rule = f"make_real('{id1}')"

        # session.commit()
        # session.begin()

        # check if id1 exists
        eid1 = session.get(Entity, id1)
        if eid1 is None:
            raise OccurrenceMissingError(f"Error, {id1} must exist in the database")

        # check if id1 is already linked
        r1_id = (
            session.query(Link.rid)
            .filter(Link.entity == id1, Link.user == user)
            .scalar()
        )

        if r1_id is not None:
            if real_id is None:
                return r1_id
            else:
                if r1_id == real_id:
                    return r1_id
                else:
                    raise RealEntityIdChangeError(
                        f"Error, {id1} is already linked to {r1_id}"
                    )

        # check if this occurence was previously linked to a real entity
        real_id = cls.recover_rentity(id1, user=user, session=session)

        if real_id is None:
            if real_id_prefix is None:
                ridp = "r" + eid1.pom_class[0]  # we take firs
            else:
                ridp = real_id_prefix
            real_id = cls.generate_id(session=session)
            real_id = f"{ridp}-{real_id}"

        if description is None:
            desc = "<No description>"
            if hasattr(eid1, "description"):
                desc = eid1.description
            elif hasattr(eid1, "name"):
                desc = eid1.name
        else:
            desc = description

        r = REntity(id=real_id, user=user, description=desc, status=status)
        session.add(r)
        session.flush()
        l1 = Link(
            rid=real_id,
            entity=id1,
            user=user,
            status=status,
            rule=rule,
            source=source,
        )
        r.links.append(l1)
        session.flush()
        session.commit()
        real_id = r.id
        return r

    def attach_occurrence(
        self,
        occ_id: str,
        user="user",
        status=LinkStatus.MANUAL,
        rule=None,
        aregister=None,
    ) -> "REntity":
        """Attach an occurrence to a real entity

        Args:
            occ_id: id of the occurrence
            user: user that linked the occurrence
            status: status of the link
            rule: rule used to link the occurrence
            aregister: id of the source of the link (normally an identifications source)

        """
        if rule is None:
            rule = f"attach({self.id},'{occ_id}')"
        with object_session(self) as session:
            # check if the entiy exists
            if session.get(Entity, occ_id) is None:
                raise OccurrenceMissingError(
                    f"Error, {occ_id} must exist in the database"
                )

            existing = [
                link
                for link in self.links
                if link.rid == self.id and link.entity == occ_id
            ]
            if len(existing) == 0:
                link = Link(
                    rid=self.id,
                    entity=occ_id,
                    user=user,
                    status=status,
                    rule=rule,
                    aregister=aregister,
                )
                self.links.append(link)
                session.merge(self)
                session.commit()
        return self

    def is_inbound_relation(self, relation):
        """Check if the relation is inbound to this real entity."""
        return relation.destination in self.get_occurrences()


class Link(Base):
    """
    Represents the link between a real entity and an entity.

    The link between a real entity and an entity is maintained in this table.

    Fields:
        rid: id of the real entity
        entity: id of the entity (occurrence)
        rule: rule used to link the entity to the real entity
        user: user that linked the entity to the real entity
        status: status of the link
        source: id of source of the link when same_as or x_same_as
        aregister: id of source of the link when same_as or x_same
    """

    __tablename__ = "links"

    __table_args__ = (UniqueConstraint("rid", "entity", "user", name="unique_link"),)

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    rid: Mapped[str] = mapped_column(
        String(64), ForeignKey(REntity.id, ondelete="CASCADE"), index=True
    )
    entity: Mapped[str] = mapped_column(
        String(64),
        ForeignKey(Entity.id),
        nullable=True,
        index=True,
    )
    # Define the relationship with Entity
    entity_rel = relationship("Entity", back_populates="links", passive_deletes=True)

    rentity_rel: Mapped[REntity] = relationship(
        REntity, back_populates="links", foreign_keys=[rid]
    )  # Many-to-one relationship Many side

    # The next two are redundant de normalized for efficiency
    user: Mapped[str] = mapped_column(
        String(64)
    )  # user that linked the entity to the real entity
    rule: Mapped[str] = mapped_column(
        String(4096)
    )  # rule used to link the entity to the real entity

    status: Mapped[LinkStatus] = mapped_column(
        Enum(LinkStatus), nullable=False
    )  # status of the link
    source: Mapped[Optional[str]] = mapped_column(
        String(64), ForeignKey(Source.id, ondelete="CASCADE"), index=True
    )  # id of source of the link when same_as or x_same_as

    aregister: Mapped[Optional[str]] = mapped_column(
        String(64), ForeignKey(ARegister.id, ondelete="CASCADE")
    )  # id of source of the link when it comes from an aregister

    @classmethod
    def missing_occurrences(cls, session=None):
        """
        Find link rows where the entity is missing in the Entity table

        """
        if session is None:
            raise ValueError("Error, session needed")
        missing = (
            session.query(Link)
            .filter(~Link.entity.in_(session.query(Entity.id)))
            .all()
        )
        return missing

    def __repr__(self):
        return f"Link(rid={self.rid}, entity={self.entity}, rule={self.rule})"


class BLink(Base):
    """Table to store backups of the links table.

    This is used when links are removed temporalily so that
    new links with the same occurrences can recover a previous
    real id. Check REntity.same_as for more details.
    """

    __tablename__ = "blinks"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    rid: Mapped[str] = mapped_column(
        String(64), ForeignKey(REntity.id, ondelete="CASCADE"), index=True
    )
    entity: Mapped[str] = mapped_column(
        String(64),  # we do not use ForeignKey because the entity may not exist
        nullable=True,
        index=True,
    )

    # The next two are redundant de normalized for efficiency
    user: Mapped[str] = mapped_column(
        String(64)
    )  # user that linked the entity to the real entity

    source: Mapped[Optional[str]] = mapped_column(
        String(64), index=True  # no foreign key because the source may not exist
    )  # id of source of the link when same_as or x_same_as

    rule: Mapped[str] = mapped_column(
        String(4096)
    )  # rule used to link the entity to the real entity

    status: Mapped[LinkStatus] = mapped_column(
        Enum(LinkStatus), nullable=False
    )  # status of the link

    aregister: Mapped[Optional[str]] = mapped_column(
        String(64),  # no foreign key because the aregister may not exist
    )  # id of source of the link when it comes from an aregister



import logging
from datetime import datetime, timezone
from enum import Enum
from typing import List
import copy

from sqlalchemy import text, delete
from sqlalchemy.orm import Session
from sqlalchemy.exc import IntegrityError

from timelink.api.models.pom_som_mapper import PomSomMapper as PomSomMapperTL
from timelink.api.models.pom_som_mapper import (
    PomClassAttributes as PomClassAttributesTL,
)
from timelink.api.models.base_mappings import (
    pom_som_base_mappings as pom_som_base_mappingsTL,
)
from timelink.api.models.base import (
    Entity as EntityTL,
    Person as PersonTL,
    Relation as RelationTL
)
from timelink.api.models.rentity import (
    REntity as REntityTL,
    Link as LinkTL,
    BLink as BLinkTL
)

from timelink.api.models.base import REntity, LinkStatus as STATUS
from timelink.api.models.system import KleioImportedFile as KleioFileTL


from timelink.mhk.models.db import pom_som_base_mappings as pom_som_base_mappingsMHK
from timelink.mhk.models.pom_som_mapper import (
    PomSomMapper as PomSomMapperMHK,
    PomClassAttributes as PomClassAttributesMHK,
)
from timelink.mhk.models.entity import Entity as EntityMHK
from timelink.mhk.models.person import Person as PersonMHK
from timelink.mhk.models.relation import Relation as RelationMHK
from timelink.mhk.models.system import KleioImportedFile as KleioFileMHK

from timelink.kleio.groups import KGroup


class KleioContext(Enum):
    """Kleio context enumeration"""

    START = (1,)
    KLEIO = (2,)
    CLASS = (3,)
    GROUP = (4,)
    ELEMENT = (5,)
    CORE = (6,)
    COMMENT = 7
    ORIGINAL = 8


class KleioHandler:
    """This class handles the import of data from a Kleio XML file into a Timelink database.

    It implements an interface to the SAX parser that reads the XML file and
    generates Kleio events. The KleioHandler class processes these events and
    stores the data in the Timelink database.

    The events are:
        newKleioFile:  start of a new Kleio file
        newClass:  definition of a new class (mapping between SOM and POM)
        newGroup:  a new kleio Group
        newRelation:  a new meta relation between kleio groups such as same_as
    """

    pom_som_base_mappings = None
    pom_som_mapper = None
    pom_class_attributes = None
    entity_model = None
    person_model = None
    kleio_file_model = None
    model_type: str = None
    postponed_relations = []
    sources_in_file = []
    xrelations = {}
    xlinks = {}
    errors = []
    warnings = []
    kleio_file_origin = None
    kleio_file = None
    kleio_file_name = None
    kleio_structure = None
    kleio_translator = None
    kleio_when = None
    kleio_context = None
    pom_som_cache = dict()
    kleio_file_is_aregister = False

    def __init__(self, session: Session, mode="TL", user="user"):
        """
        Arguments:
            session: a SQLAlchemy session
            mode: the mode of the import, either 'TL'(Timelink) or 'MHK'
            user: the user that is importing the data, used for same as ownership

        """
        self.session = session
        self.user = user
        self.kleio_file_is_aregister = False
        self.kleio_source_id = None
        self.aregister_id = None
        self.xrefs = {}
        if mode == "TL":
            self.pom_som_base_mappings = pom_som_base_mappingsTL
            self.pom_som_mapper = PomSomMapperTL
            self.pom_class_attributes = PomClassAttributesTL
            self.pom_class_attributes = PomClassAttributesTL
            self.entity_model = EntityTL
            self.person_model = PersonTL
            self.relation_model = RelationTL
            self.rentity_model = REntityTL
            self.link_model = LinkTL
            self.kleio_file_model = KleioFileTL
            self.model_type = "TL"
        elif mode == "MHK":
            self.pom_som_base_mappings = pom_som_base_mappingsMHK
            self.pom_som_mapper = PomSomMapperMHK
            self.pom_class_attributes = PomClassAttributesMHK
            self.entity_model = EntityMHK
            self.person_model = PersonMHK
            self.relation_model = RelationMHK
            self.kleio_file_model = KleioFileMHK
            self.model_type = "MHK"
        else:
            raise ValueError(f"Unknown mode {mode}")

    def save_source_context(self, source_id):
        """Save all the links and relations pointing to this
        source from other sources because they will be nulled or deleted
        when the source is deleted.

        After import we will restore them.

        This will add to the self.cross_source_references dict

        We use the follow queries to get the external relations to this source:

        ```sql
        SELECT r.id, r.origin, r.destination, r.the_type, r.the_value
                FROM relations r, entities e
                WHERE r.id = e.id
                            AND e.the_source != :source_id
                            AND r.destination IN
                                (SELECT id
                                 FROM entities e2
                                 WHERE e2.id = r.destination
                                       e2.the_source = :source_id
                                        AND e2.the_source != e.the_source)

        ```
        And to get the external links to this source:

        ```sql
        SELECT l.rid,l.entity,l.source as link_source, e.the_source as entity_source
                FROM links l, entities e
                WHERE l.entity = e.id
                AND e.the_source = :source_id
                AND link_source != e.the_source;
        """
        # Currently not implemented in MHK Mysql databases
        if self.model_type == "MHK":
            return
        # make a sql alchemy query to get the cross source relations
        # and store them in cross_source_relations
        sql = """select r.id,
                    r.origin,
                    r.destination,
                    r.the_type,
                    r.the_value,
                    r.the_date,
                    e.the_source,
                    e2.the_source as dest_source
                        FROM relations r, entities e, entities e2
                        WHERE r.id = e.id
                            AND r.destination = e2.id
                            AND e2.the_source = :source_id
                            AND e.the_source != e2.the_source
                """
        # https://docs.sqlalchemy.org/en/20/orm/queryguide/select.html#orm-queryguide-selecting-text
        sql_query = (
            text(sql)
            .bindparams(source_id=source_id)
            .columns(
                self.relation_model.id,
                self.relation_model.origin,
                self.relation_model.destination,
                self.relation_model.the_type,
                self.relation_model.the_value,
                self.relation_model.the_date,
                self.entity_model.the_source,
                self.entity_model.the_source.label("dest_source"),
            )
        )
        xrefs = self.session.execute(sql_query).fetchall()

        self.xrelations[source_id] = xrefs

        # get the links that are going to be affected by the reimport
        sql = """
                SELECT l.id, l.rid,l.entity, l.user, l.rule, l.status, l.source as link_source, e.the_source as entity_source
                FROM links l, entities e
                WHERE l.entity = e.id
                AND e.the_source = :source_id;

        """
        sql_query = (
            text(sql)
            .bindparams(source_id=source_id)
            .columns(
                self.link_model.id,
                self.link_model.rid,
                self.link_model.entity,
                self.link_model.user,
                self.link_model.rule,
                self.link_model.status,
                self.link_model.source,
                self.entity_model.the_source.label('entity_source')
            )
        )
        xlinks = self.session.execute(sql_query).fetchall()

        # save the links that are going to be affected by the reimport
        for link_id in [link.id for link in xlinks]:
            link = self.session.get(self.link_model, link_id)

            # find a blink for the same rid, entity and user
            # if it exists we do not create a new one

            blink = self.session.query(BLinkTL).filter(
                BLinkTL.rid == link.rid,
                BLinkTL.entity == link.entity,
                BLinkTL.user == link.user
            ).first()

            if blink is None:
                # create a blink with the link information
                blink = BLinkTL(
                    rid=link.rid,
                    entity=link.entity,
                    user=link.user,
                    rule=link.rule,
                    status=link.status,
                    source=link.source,
                )
            self.session.add(blink)
            self.session.commit()

        # maybe not useful
        self.xlinks[source_id] = xlinks

    def restore_source_context(self, source_id):
        """Restore all the links and relations pointing to this
        source from other sources that were saved before the source was deleted

        TODO: is this better here or in Source entity model?
        Reasons for here: at the import level restore only
        makes sense after all the source is imported because
        at the entity level a Source is inserted empty and only
        at the end of the source import it is possible to restore
        the missing links
        """
        if self.model_type == "MHK":
            return
        relations = self.xrelations[source_id]
        for r in relations:
            # we get the relation in other sources that was affect by the reimport
            rel = self.session.get(
                self.relation_model, r.id
            )  # safe because id is on the exernal source
            if rel is not None:
                # this relation should have the destination None
                # because the destination was deleted when the current source
                # was deleted before reimport
                # check that the old destination still exists after reimport
                if rel.destination is not None:
                    # if the relation has a destination it means that it was not affected
                    # by the reimport of the source, but it should
                    # show a warning
                    self.warnings.append(
                        f"WARNING: An external relation {r.id} "
                        f"of type {r.the_type} with value {r.the_value} "
                        f"referred entity {r.destination} in source {r.dest_source} "
                        f"was not affected by reimport of source {source_id}"
                    )
                dest = self.session.get(self.entity_model, r.destination)
                if dest is not None:
                    # if so we restore the relation destination
                    rel.destination = r.destination  # set to the saved value
                    self.session.merge(rel)
                    self.session.commit()

                    # TODO: and now we should process the xsame_as relations
                    if rel.the_type == 'identification' and rel.the_value == 'same as':

                        REntity.same_as(
                            rel.origin,
                            rel.destination,
                            status=STATUS.SOURCE,
                            rule=f"same_as('{self.kleio_file_name}', 'reimported')",
                            source=source_id,
                            session=self.session,
                        )
                        pass
                else:
                    # if the destination does not exist
                    # we flag the error
                    self.errors.append(
                        f"ERRROR: An external relation {r.id} "
                        f"of type {r.the_type} with value {r.the_value} "
                        f"referred entity {r.destination} which does not exist "
                        f"after reimport of source {source_id}"
                    )
                self.session.commit()
            else:
                # The saved relation in the other source was deleted
                # this should not have happened
                self.errors.append(
                    f"ERRROR: An external relation {r.id} "
                    f"of type {r.the_type} with value {r.the_value} "
                    f"with origin {r.origin} and destination {r.destination}"
                    f"was saved before reimport of source {source_id} "
                    f"but was deleted during reimport."
                )

    def newKleioFile(self, attrs):
        """Process a new Kleio file
        #<KLEIO STRUCTURE="/kleio-home/system/conf/kleio/stru/gacto2.str"
                SOURCE="/kleio-home/sources/soure-fontes/sources/1685-1720/baptismos/b1685.cli"
                TRANSLATOR="gactoxml2.str"
                WHEN="2020-8-18 17:10:32"
                OBS="" SPACE="">
        # TODO: should use a subclass for this
        """
        self.session.commit()
        self.postponed_relations = []
        self.errors = []
        self.warnings = []
        self.kleio_file = attrs["SOURCE"]
        # extract prefix /kleio-home/ from self.kleio_file
        if self.kleio_file.startswith("/kleio-home/"):
            self.kleio_file = self.kleio_file.replace("/kleio-home/", "")
        # extract name of file from path
        self.kleio_file_name = self.kleio_file.split("/")[-1]
        self.kleio_structure = attrs["STRUCTURE"]
        if self.kleio_structure.startswith("/kleio-home/"):
            self.kleio_structure = self.kleio_structure.replace("/kleio-home/", "")
        self.kleio_translator = attrs["TRANSLATOR"]
        self.kleio_when = attrs["WHEN"]
        self.kleio_file_is_aregister = False
        self.kleio_source_id = None
        self.aregister_id = None
        self.pom_som_cache = dict()
        self.sources_in_file = []
        self.xrefs = {}
        self.xlinks = {}
        logging.debug("Kleio file: %s", self.kleio_file)
        logging.debug("Kleio structure: %s", self.kleio_structure)
        logging.debug("Kleio translator: %s", self.kleio_translator)
        logging.debug("Kleio when: %s", self.kleio_when)
        # mark that file is being imported by nulling the imported date
        kfile = self.session.get(self.kleio_file_model, self.kleio_file)
        if kfile is not None:
            kfile.imported = None
            kfile.imported_string = None
            self.session.commit()

    def newClass(
        self,
        psm: PomSomMapperTL | PomSomMapperMHK,
        attrs: List[PomClassAttributesTL | PomClassAttributesMHK],
    ):
        if psm.id in self.pom_som_base_mappings.keys():
            # we do not allow redefining of base mappings
            # TODO needs rethinking see #53
            logging.debug("Skipping base mapping %s", psm.id)
            orm_class = self.entity_model.get_orm_for_pom_class(psm.id)
            if orm_class is not None:
                self.pom_som_mapper.group_orm_models[psm.group_name] = orm_class
            return

        # if we import mapping from local file we also skip those
        # the importer should check at the beggining if local mappings
        # are available before starting the file import
        # we need a function check_local_mappings() that returns a list
        # of mappings available in the local file system.
        # if psm.id in check_local_mappings(): return

        session = self.session

        # check if we have this class defined in the database.
        existing_psm: PomSomMapperTL | PomSomMapperMHK | None = session.get(
            self.pom_som_mapper, psm.id
        )
        if existing_psm is not None:  # class exists: delete, insert again
            try:
                logging.debug("Deleting existing class %s", psm.id)
                session.commit()
                stmt = delete(self.pom_class_attributes).where(
                    self.pom_class_attributes.the_class == psm.id
                )
                session.execute(stmt)
                session.delete(existing_psm)
                session.commit()
            except Exception as e:
                session.rollback()
                self.errors.append(
                    f"ERROR: deleting class {psm.id}: {e.__class__.__name__}: {e}"
                )

        # now we add the new mapping
        try:
            logging.debug("Adding class %s", psm.id)
            session.add(psm)
            for class_attr in attrs:
                session.add(class_attr)
            session.commit()
        except Exception as e:
            session.rollback()
            self.errors.append(
                f"ERROR: adding class {psm.id}: {e.__class__.__name__}: {e}"
            )
        # ensure that the table and ORM classes are created
        try:
            psm.ensure_mapping(session)
            session.commit()
        except Exception as e:
            session.rollback()
            self.errors.append(
                f"ERROR: creating ORM mapping for class {psm.id}: {e.__class__.__name__}: {e}"
            )

    def newGroup(self, group: KGroup):
        # get the PomSomMapper
        # pass storeKGroup
        pom_mapper_for_group: PomSomMapperTL | PomSomMapperMHK
        try:
            if group.pom_class_id in self.pom_som_cache.keys():
                pom_mapper_for_group = self.pom_som_cache[group.pom_class_id]
            else:
                pom_mapper_for_group = self.pom_som_mapper.get_pom_class(
                    group.pom_class_id, self.session
                )
            if pom_mapper_for_group is None:
                raise ValueError(
                    f"Could not find PomSomMapper for class {group.pom_class_id}"
                )
            self.pom_som_cache[group.pom_class_id] = (
                pom_mapper_for_group  # TODO redundant?
            )
        except Exception as exc:
            self.errors.append(
                f"ERROR: {self.kleio_file_name} {str(group.line)}"
                f" finding PomSomMapper for class "
                f"{group.pom_class_id}: {exc.__class__.__name__}: {exc}"
            )
            return

        # store the current source id into the group
        group.source_id = self.kleio_source_id

        # check if pom_mapper_for_group is attached to the session
        if not self.session.object_session(pom_mapper_for_group):
            self.session.add(pom_mapper_for_group)

        if group.pom_class_id == "act":  # TODO should be .extends("act")
            logging.debug(f"{group._name}${group.id}")

        if (
            pom_mapper_for_group.id == "aregister"
        ):  # TODO should be .extends("aregister")
            self.kleio_file_is_aregister = True  # current file is an authority register
            self.aregister_id = str(group.id.core)

        if pom_mapper_for_group.id == "source":  # TODO should be .extends("source")
            self.kleio_source_id = str(group.id.core)
            group.source_id = self.kleio_source_id
            self.sources_in_file.append(self.kleio_source_id)
            # We need to save all the links and relations pointing to this
            # source from other sources because they will be nulled or deleted
            # when the source is deleted
            # after import we will restore them for the entities that are restored
            # in this import
            self.save_source_context(self.kleio_source_id)

        if pom_mapper_for_group.id == "relation":  # TODO should be .extends("relation")
            # it can happen that the destination of a relation
            #  is not yet in the database (forward reference in relation)
            #  In this case we postpone the storing of the relation
            # until the end of file
            exits_dest_rel = self.session.get(
                self.entity_model,
                group.get_element_by_name_or_class("destination").core,
            )
            if exits_dest_rel is None:
                self.postponed_relations.append(
                    (pom_mapper_for_group.id, copy.deepcopy(group))
                )
            else:
                try:
                    pom_mapper_for_group.store_KGroup(group, self.session)
                except Exception as exc:
                    self.session.rollback()
                    self.errors.append(
                        f"ERROR: {self.kleio_file_name} {str(group.line)} "
                        f"storing group {group.kname}${group.id}: "
                        f"{exc.__class__.__name__}: {exc}"
                    )
                    self.session.rollback()
        else:
            # we have the POM class, less check special cases
            try:
                pom_mapper_for_group.store_KGroup(group, self.session)
            except IntegrityError as ierror:
                logging.error(
                    f"ERROR: {self.kleio_file_name} line {str(group.line)} "
                    f"integrity error {group.kname}${group.id}: {ierror}",
                    stacklevel=3
                )
                self.errors.append(
                    f"ERROR: {self.kleio_file_name} line {str(group.line)} "
                    f"** integrity error {group.kname}${group.id}: {ierror}"
                )
                self.session.rollback()

            except Exception as exc:
                logging.error(
                    f"ERROR: {self.kleio_file_name} line {str(group.line)} "
                    f"storing group {group.kname}${group.id}: {exc.__class__.__name__}: {exc}",
                    stacklevel=3
                )
                self.session.rollback()
                self.errors.append(
                    f"ERROR: {self.kleio_file_name} line {str(group.line)} "
                    f"storing group {group.kname}${group.id}: {exc.__class__.__name__}: {exc}"
                )

    def newRelation(self, attrs):
        """Process a new (Meta) relation

        Currently there are two types of metadata relations,
        both related to processing same as relations:

        1. same_as:  register the same_as and xsame_as elements
                    in groups
        ```xml
        <RELATION ID="__-sa-deh-joao-barradas-rela6"
            ORG="sa-deh-joao-barradas-per1-11"
            DEST="sa-deh-matteo-ricci"
            TYPE="META"
            VALUE="same_as"/>
        ```

        2. attach_to_rentity:  used in authority registers to register
                                existing real entities to the register
        ```xml
            <RELATION ID="rp-31-occ1"
                    REGISTER="real-entities-toliveira"
                    USER="toliveira"
                    ORG="deh-paul-lieou"
                    DEST="rp-31"
                    TYPE="META"
                    VALUE="attach_to_rentity" />
        ```
        """
        rel_id = attrs["ID"]
        rel_origin = attrs["ORG"]
        rel_dest = attrs["DEST"]
        rel_value = attrs["VALUE"]
        rel_register = attrs.get("REGISTER", None)
        rel_user = attrs.get("USER", None)

        if rel_value == "same_as":
            self.session.commit()

            # register the same_as relation

            try:
                REntity.same_as(
                    rel_origin,
                    rel_dest,
                    status=STATUS.SOURCE,
                    rule=f"same_as('{self.kleio_file_name}')",
                    source=self.kleio_source_id,
                    session=self.session,
                )
                self.session.commit()
            except Exception as exc:
                self.errors.append(
                    f"ERROR: {self.kleio_file_name} "
                    f"processing same_as relation {rel_id}: {exc.__class__.__name__}: {exc}"
                )
                self.session.rollback()
                return

        elif rel_value == "attach_to_rentity":
            # register the attach_to_rentity relation
            self.session.commit()
            rentity = self.session.get(REntity, rel_dest)
            if rentity is None:
                error_msg = (
                    f"ERROR: {self.kleio_file_name} "
                    f"processing attach_to_rentity relation {rel_id}: "
                    f"Real entity {rel_dest} not found"
                )
                logging.error(error_msg)
                self.errors.append(error_msg)
                self.session.rollback()
                return
            try:
                rentity.attach_occurrence(
                    occ_id=rel_origin,
                    user=rel_user,
                    status=rentity.status,
                    rule=f"attach_to_rentity('{rel_register}')",
                    aregister=self.aregister_id,
                )
                self.session.commit()
            except Exception as exc:
                msg = (
                    f"ERROR: {self.kleio_file_name} "
                    f"processing attach_to_rentity relation {rel_id}: {exc.__class__.__name__}: {exc}"
                )
                logging.error(msg)
                self.errors.append(msg)
                self.session.rollback()
                return

        else:  # unknown relation
            self.errors.append(
                f"ERROR: {self.kleio_file_name} " f"unknown relation type {rel_value}"
            )
            self.session.rollback()

    def endKleioFile(self):
        """Process end of file: process postponed relations"""
        # store postponed relations
        postponed = len(self.postponed_relations)
        if postponed > 0:
            log = f"Storing {postponed} postponed relations"
            logging.info(log)
        for pom_mapper_id, group in self.postponed_relations:
            pom_mapper = self.pom_som_mapper.get_pom_class(pom_mapper_id, self.session)
            try:
                pom_mapper.store_KGroup(group, self.session)
                self.session.commit()
            except IntegrityError as ierror:
                self.errors.append(
                    f"ERROR: {self.kleio_file_name}: {str(group.line)} "
                    f"storing {group.to_kleio()}\n {ierror}"
                )
                self.session.rollback()
            except Exception as exc:
                self.errors.append(
                    f"ERROR: {self.kleio_file_name} {str(group.line)} "
                    f"storing {group.to_kleio()}\n {exc.__class__.__name__}: {exc}"
                )
                self.session.rollback()
        self.postponed_relations = []

        # restore all the links and relations pointing to this
        # source(s) from other sources that were saved before the source was deleted
        # the cross references were save when a source group was processed
        # TODO: should this go here or to models.Source?
        for source_id in self.sources_in_file:
            self.restore_source_context(source_id)

        # store information on kleio file import
        kfile = self.kleio_file_model(
            path=self.kleio_file,
            name=self.kleio_file_name,
            structure=self.kleio_structure,
            translator=self.kleio_translator,
            # convert date from "2020-8-18 17:10:32" format
            # to datetime object
            translation_date=datetime.strptime(self.kleio_when, "%Y-%m-%d %H:%M:%S"),
            nerrors=len(self.errors),
            nwarnings=len(self.warnings),
        )
        if len(self.errors) > 0:
            s = "\n\n".join(self.errors)
        else:
            s = "No errors"
        kfile.error_rpt = s
        if len(self.warnings) > 0:
            s = "\n\n".join(self.warnings)
        else:
            s = "No warnings"
        kfile.warning_rpt = s
        kfile.imported = datetime.now(timezone.utc)
        kfile.imported_string = datetime.now(timezone.utc).strftime(
            "%Y-%m-%d %H:%M:%S %Z"
        )
        kfile_exists = self.session.get(self.kleio_file_model, kfile.path)
        if kfile_exists is not None:
            self.session.delete(kfile_exists)
            self.session.commit()
        self.session.add(kfile)
        self.session.commit()



from timelink.kleio.groups.kgroup import KGroup
from timelink.kleio.groups.kact import KAct


class KGeoentity(KGroup):
    """KGeoentity(name,type,id=,obs=,same_as=,xsame_as=)

    A geographic entity. Flags an Entity as a geographic entity.
    No specific fields are defined for this class.

    Elements:
        name: the name of the object. A string.
        type:  the . A string.
        id: an unique id for this person. A string, optional.
        obs: a note on the person. A string, optional.
        same_as: id of another instance of this object in the same file.
        xsame_as: id of another instance of this object in another file.

    Kleio str definition:

            part name=geoentity;
                position=name; guaranteed=name;
                also=type,id,obs,same_as,xsame_as;
                part=ls,atr,rel

    """

    _name = "geoentity"
    _guaranteed = ["name"]
    _also = ["id", "type", "obs", "same_as", "xsame_as"]
    _position = ["name", "id"]
    _part = ["rel", "attr"]
    _pom_class_id: str = "geoentity"


KAct.allow_as_part(KGeoentity)



""" Base class, all modules must import this to share Metadata

   for 2.0 migration see:
   https://docs.sqlalchemy.org/en/20/changelog/whatsnew_20.html#whatsnew-20-orm-declarative-typing

"""

from typing import Any
from sqlalchemy import MetaData  # pylint: disable=import-error
from sqlalchemy.orm import DeclarativeBase  # pylint: disable=import-error
from sqlalchemy.orm import declared_attr  # pylint: disable=import-error


naming_convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_`%(constraint_name)s`",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


class Base(DeclarativeBase):
    """Base class, all modules must import this to share Metadata"""

    # see https://alembic.sqlalchemy.org/en/latest/naming.html#naming-conventions
    metadata = MetaData(naming_convention=naming_convention)

    # see https://docs.sqlalchemy.org/en/20/changelog/migration_20.html#migration-20-step-six
    __allow_unmapped__: bool = True

    id: Any
    __name__: str

    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(cls) -> str:  # pylint: disable=no-self-argument
        return cls.__name__.lower()



from timelink.api.database import get_postgres_dbnames
from tests import skip_on_travis

# Skip tests if running on Travis CI
pytestmark = skip_on_travis


def test_get_dbnames():
    dbnames = get_postgres_dbnames()
    assert len(dbnames) > 0, "No databases found"



""" Timelink User model

timelink users are registered users of the application.

Each user can be associated with one or more "projects".

A project is a collection of Kleio sources and a database to store the data.

Access of the user to Kleio sources is done through a Kleio server.
In the Kleio server each user has an acess token that determines the
access level to the sources directory. So for each project the user
needs to have a valid access token to the Kleio server and its url.

Each project is associated with a database. Timelink is tested with
PostgreSQL and SQLite databases.

When using SQLite, the database is a file in the file system.
When using PostgreSQL, each project is associated with a database
(see https://www.postgresql.org/docs/8.0/managing-databases.html)
and each project is associated with a user and a password to access the database.

users have different access levels to the projects. The access levels are:

1. admin: the user has full access to the project, including the ability to associate
    new users to the project.
2. manager: the user has access to the project, can import data and identify
    entities in the data, can add commentaries, but cannot associate new users
    to the project.
3. colaborator: the user has access to the project, cannot import data, but can
    identify entities in the data and add commentaries.
4. viewer: the user has access to the project, but cannot import data, identify
    people or add commentaries.

"""

from datetime import datetime
from typing import List, Optional
from sqlalchemy import String
from sqlalchemy.orm import DeclarativeBase

from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy import DateTime
from sqlalchemy.sql import func
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship

from timelink.app.schemas.user import UserPropertySchema, UserSchema


# declarative base class
class Base(DeclarativeBase):
    pass


class UserProperty(Base):
    __tablename__ = "user_properties"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(30))
    value: Mapped[str] = mapped_column(String(30))
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"))

    user: Mapped["User"] = relationship("User", back_populates="properties")

    def as_schema(self):
        """Return the pydantic model for this object."""
        return UserPropertySchema.model_validate(self)

    def __repr__(self):
        return f"UserProperty(id={self.id}, name={self.name}, value={self.value}, user_id={self.user_id})"

    def __str__(self):
        return f"UserProperty(id={self.id}, name={self.name}, value={self.value}, user_id={self.user_id})"


class User(Base):
    """
    Attributes:
        id (int): Primary key, autoincrement.
        name (str): Required.
        nickname (str): Optional.
        email (str): Required, unique.
        hashed_password (str): Optional.
        disabled (bool): Optional, default is False.
        created (datetime): Required, default is now.
        updated (datetime): Optional, default is now, updated at now.
    """

    __tablename__ = "users"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(String(128), unique=False)
    nickname: Mapped[Optional[str]] = mapped_column(String(32))
    email: Mapped[str] = mapped_column(unique=True)
    # deprecated. Authentica with Fief
    hashed_password: Mapped[Optional[str]]
    disabled: Mapped[Optional[bool]] = mapped_column(default=False)
    created: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), default=func.now()
    )
    updated: Mapped[Optional[datetime]] = mapped_column(
        DateTime(timezone=True), default=func.now(), onupdate=func.now()
    )

    properties: Mapped[List["UserProperty"]] = relationship(
        "UserProperty", back_populates="user", cascade="all, delete, delete-orphan"
    )

    projects: Mapped[List["ProjectAccess"]] = relationship(  # type: ignore # noqa F821
        "ProjectAccess", back_populates="user", cascade="all, delete, delete-orphan"
    )

    def as_schema(self):
        """Return the pydantic model for this object."""
        us = UserSchema.model_validate(self)
        us.properties = [p.as_schema() for p in self.properties]
        us.projects = [p.as_schema() for p in self.projects]

    def __repr__(self):
        return (
            f"User("
            f"id={self.id}, "
            f"name={self.name}, "
            f"email={self.email}, "
            f"nickname={self.nickname}, "
            f"created={self.created}, "
            f"updated={self.updated})"
        )

    def __str__(self):
        sproperties = ", ".join([str(p) for p in self.properties])
        return (
            f"User("
            f"id={self.id}, "
            f"name={self.name}, "
            f"email={self.email}, "
            f"fullname={self.fullname}, "
            f"nickname={self.nickname}, "
            f"created={self.created}, "
            f"updated={self.updated}, "
            f"properties={sproperties})"
        )



from timelink.kleio.groups.kgroup import KGroup
from timelink.kleio.groups.kkleio import KKleio


class KSource(KGroup):
    """Represents an Historical Source. Sources contain :class:`KAct` and
    may contain :class:`KAttribute`.

    Elements

    id
        An unique id for this source.
    type
        The type of the source (e.g. baptisms, marriages); optional.
    loc
        Location (name of archive, library, collection); optional.
    ref
        The call reference ("cota") of the source in the location; optional.
    date
        The date of the source. A string in timelink format; optional.

          - 1582
          - 1582-05-04
          - 1582:1609
          - >1582:<1702
    year
        A single year. A number. Deprecated, use date instead
    obs
        Observations on the source (can be long and multiline); optional.
    replace
        Id of source to be replaced. A string; optional.
        The source with this id is removed before importing this one.
        Used when changing the id of a file, old id should go here.

    Kleio str definition::

        part name=historical-source;
             guaranteed=id;
             also=type,date,year,loc,ref,obs,replace;
             position=id,year,type,ref;
             part=historical-act


    """

    _name = "source"
    _guaranteed = ["id"]
    _also = ["type", "date", "year", "loc", "ref", "replace", "obs"]
    _position = ["id"]
    _part = ["act", "attr"]
    _pom_class_id: str = "source"


KKleio.allow_as_part(KSource)



"""

"""

from sqlalchemy import select
import pandas as pd

from timelink.api.database import TimelinkDatabase
from timelink.api.models import Person


def remove_particles(name, particles=None):
    if particles is None:
        particles = ("de", "da", "e", "das", "dos", "do")

    return " ".join([n for n in name.split() if n not in particles])


def pname_to_df(
    name,
    db: TimelinkDatabase = None,
    session=None,
    similar=False,
    name_particles=None,
    sql_echo=False,
):
    """name_to_df return df of people with a matching name

    Args:
        name: name to search for
        db: = database connection to use, either db or session must be specified
        session: session to use, either db or session must be specified
        similar: if true will strip particles and insert a wild card %
                between name components with an extra one at the end
        name_particles: list, particles to remove before comparing names
    """
    if db is not None:  # try if we have a db connection in the parameters
        my_session = db.session()
    elif session is not None:
        my_session = session
    else:  # no session or db connection specified
        raise (
            Exception("must provide database connection (db=) or session (session=)")
        )

    if similar:
        name_particles = remove_particles(name).split(" ")
        name_like = "%".join(name_particles)
        name_like = name_like + "%"
    else:
        name_like = name

    ptable = Person.__table__

    stmt = select(ptable.c.id, ptable.c.name, ptable.c.sex, ptable.c.obs).where(
        ptable.c.name.like(name_like)
    )

    if sql_echo:
        print(stmt)

    records = my_session.execute(stmt)
    df = pd.DataFrame.from_records(
        records.columns("id", "name", "sex", "obs"),
        index=["id"],
        columns=["id", "name", "sex", "obs"],
    )
    if df.iloc[0].count() == 0:
        df = None  # nothing found we
    else:
        df.reset_index(inplace=True)
    return df



from __future__ import annotations as _annotations

from typing import Annotated, Literal, TypeAlias
import asyncio

from fastapi import APIRouter, Depends, Request

from fastui import AnyComponent, FastUI
from fastui import components as c
from fastui.auth import AuthRedirect, GitHubAuthProvider
from fastui.events import AuthEvent, GoToEvent, PageEvent

# from httpx import AsyncClient
from pydantic import BaseModel, EmailStr, Field, SecretStr

from timelink.app.schemas.user import UserSchema
from timelink.app.dependencies import get_current_user, get_github_auth
from timelink.app.web.home_page import home_page


router = APIRouter(tags=["fastui"], responses={404: {"description": "Not found"}})


# --------------------------


UserDep = Annotated[UserSchema, Depends(get_current_user)]

ACCESS_TOKEN_EXPIRE_MINUTES = 5 * 24 * 60  # 5 days

LoginKind: TypeAlias = Literal["password", "github"]


@router.get("/login/{kind}", response_model=FastUI, response_model_exclude_none=True)
def auth_login(
    kind: LoginKind,
    request: Request,
    user: UserSchema = None,
    # user: Annotated[UserSchema, Depends(get_current_user)]
) -> list[AnyComponent]:
    if user is not None:
        # already logged in
        raise AuthRedirect("/")

    return home_page(
        c.LinkList(
            links=[
                c.Link(
                    components=[c.Text(text="Password Login")],
                    on_click=PageEvent(
                        name="tab",
                        push_path="/auth/login/password",
                        context={"kind": "password"},
                    ),
                    active="/auth/login/password",
                ),
                c.Link(
                    components=[c.Text(text="GitHub Login")],
                    on_click=PageEvent(
                        name="tab",
                        push_path="/auth/login/github",
                        context={"kind": "github"},
                    ),
                    active="/auth/login/github",
                ),
            ],
            mode="tabs",
            class_name="+ mb-4",
        ),
        c.ServerLoad(
            path="/auth/login/content/{kind}",
            load_trigger=PageEvent(name="tab"),
            components=auth_login_content(kind),
        ),
        request=request,
        title="Authentication",
        user=user,
    )


@router.get(
    "/login/content/{kind}", response_model=FastUI, response_model_exclude_none=True
)
def auth_login_content(kind: LoginKind) -> list[AnyComponent]:
    match kind:
        case "password":
            return [
                c.Heading(text="Timelink login", level=5, class_name="text-center"),
                c.ModelForm(
                    model=LoginForm,
                    submit_url="/fastui/auth/login",
                    display_mode="page",
                ),
            ]
        case "github":
            return [
                c.Heading(text="GitHub Login", level=3),
                c.Paragraph(text="Demo of GitHub authentication."),
                c.Paragraph(
                    text="(Credentials are stored in the browser via a JWT only)"
                ),
                c.Button(
                    text="Login with GitHub",
                    on_click=GoToEvent(url="/auth/login/github/gen"),
                ),
            ]
        case _:
            raise ValueError(f"Invalid kind {kind!r}")


class LoginForm(BaseModel):
    email: EmailStr = Field(
        title="Email Address",
        description="Enter a valid email address",
        json_schema_extra={"autocomplete": "email"},
    )
    password: SecretStr = Field(
        title="Password",
        description="Enter your password",
        json_schema_extra={"autocomplete": "current-password"},
    )


@router.get("/profile", response_model=FastUI, response_model_exclude_none=True)
async def profile(
    request: Request, user: Annotated[UserSchema, Depends(get_current_user)]
) -> list[AnyComponent]:
    return home_page(
        c.Paragraph(text=f'You are logged in as "{user.email}".'),
        c.Button(text="Logout", on_click=PageEvent(name="submit-form")),
        c.Heading(text="User Data:", level=3),
        c.Code(language="json", text=user.model_dump_json(indent=2)),
        c.Form(
            submit_url="/fastui/auth/logout",
            form_fields=[
                c.FormFieldInput(
                    name="test", title="", initial="data", html_type="hidden"
                )
            ],
            footer=[],
            submit_trigger=PageEvent(name="submit-form"),
        ),
        title="Authentication",
        request=request,
        user=user,
    )


@router.post("/logout", response_model=FastUI, response_model_exclude_none=True)
async def logout_form_post() -> list[AnyComponent]:
    return [c.FireEvent(event=AuthEvent(token=False, url="/auth/login/password"))]


@router.get(
    "/login/github/gen", response_model=FastUI, response_model_exclude_none=True
)
async def auth_github_gen(
    github_auth: Annotated[GitHubAuthProvider, Depends(get_github_auth)]
) -> list[AnyComponent]:
    auth_url = await github_auth.authorization_url()
    return [c.FireEvent(event=GoToEvent(url=auth_url))]


@router.get(
    "/login/github/redirect", response_model=FastUI, response_model_exclude_none=True
)
async def github_redirect(
    code: str,
    state: str | None,
    github_auth: Annotated[GitHubAuthProvider, Depends(get_github_auth)],
) -> list[AnyComponent]:
    exchange = await github_auth.exchange_code(code, state)
    user_info, emails = await asyncio.gather(
        github_auth.get_github_user(exchange),
        github_auth.get_github_user_emails(exchange),
    )
    user = UserSchema(
        email=next((e.email for e in emails if e.primary and e.verified), None),
        extra={
            "github_user_info": user_info.model_dump(),
            "github_emails": [e.model_dump() for e in emails],
        },
    )
    token = user.encode_token()
    return [c.FireEvent(event=AuthEvent(token=token, url="/auth/profile"))]



from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    """Timelink settings."""

    timelink_app_name: str = "Timelink"
    timelink_admin_email: str = "joaquim@uc.pt"
    timelink_users_db_type: str = "sqlite"
    timelink_users_db_name: str = "timelink_users.sqlite"
    timelink_app_url: str = "http://localhost:8008"
    timelink_users_db_path: str = ""
    timelink_admin_pwd: str = "adminpwd"
    timelink_project_db_type: str = "sqlite"



"""Add index to entities

Revision ID: 6ccf1ef385a6
Revises: 36cdd078f890
Create Date: 2024-12-23 19:26:13.287735

"""
from typing import Sequence, Union

from alembic import op


# revision identifiers, used by Alembic.
revision: str = '6ccf1ef385a6'
down_revision: Union[str, None] = '36cdd078f890'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('entities_class_group', 'entities', ['class', 'groupname'], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('entities_class_group', table_name='entities')

    # ### end Alembic commands ###



from timelink.kleio.groups.kobject import KObject
from timelink.kleio.groups.kact import KAct


class KAbstraction(KObject):
    """A synonym for object, used in non physical entities such as
    institutions.

    """

    pass


KAct.allow_as_part(KAbstraction)



""" Test timelink Pandas integrations"""

import pytest
from pathlib import Path
# import pdb  # noqa  TODO: remove when no longer needed

from sqlalchemy_utils import drop_database  # pylint: disable=unused-import. # noqa

from timelink.api.database import TimelinkDatabase
from timelink.api.models import base  # pylint: disable=unused-import. # noqa: F401
from tests import TEST_DIR

from timelink.pandas import pname_to_df, entities_with_attribute, attribute_values

# pytestmark = skip_on_travis

test_set = [("sqlite", "test_pandas_db"), ("postgres", "test_pandas_db")]
test_files = "projects/test-project/sources/reference_sources/pandas"


@pytest.fixture(scope="module")
def dbsystem(request, kleio_server):
    """Create a database for testing"""
    db_type, db_name = request.param
    db_path = Path(TEST_DIR, "sqlite")

    database = TimelinkDatabase(db_name, db_type, db_path=db_path, echo=False)
    database.set_kleio_server(kleio_server)
    database.update_from_sources(test_files)

    try:
        yield database
    finally:
        with database.session() as session:
            pass
            database.drop_db(session)
            session.close()
        if ":memory:" not in database.db_url:
            pass
            # drop_database(database.db_url)


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_name_to_df(dbsystem):
    """test generation of dataframe with people with a certain name"""

    # test finding by similarity
    names_df = pname_to_df("matias carvalho", db=dbsystem, similar=True, sql_echo=True)
    print(names_df)
    assert names_df is not None, "name_to_df returned None"
    assert len(names_df) > 0

    # test finding by exact match
    names_df = pname_to_df("matias carvalho", sql_echo=True, db=dbsystem, similar=False)
    assert names_df is None, "name_to_df returned results but should not have"
    print(names_df)


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_attribute_values(dbsystem):
    """test generation of dataframe with people/objects with a certain attribute"""

    locais_residencia = attribute_values("residencia", db=dbsystem, sql_echo=True)
    assert locais_residencia is not None, "attribute_values returned None"
    attr_values = attribute_values("nacionalidade", db=dbsystem, sql_echo=True)
    assert attr_values is not None, "attribute_values returned None"
    # get count of unique values for Portugal
    attr_type_count = attr_values.loc["Portugal"]["count"]
    # now count filtering by groupname "n" (this filters out "referidos"
    attr_values_2 = attribute_values(
        "nacionalidade", groupname="n", db=dbsystem, sql_echo=True
    )
    # get count of unique values for Portugal
    attr_type_count_2 = attr_values_2.loc["Portugal"]["count"]
    assert attr_type_count > attr_type_count_2, "attribute_values returned bad results"


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_entities_with_attribute(dbsystem):
    """Test generation of dataframe from attributes"""
    df = entities_with_attribute(
        db=dbsystem,
        entity_type="person",
        the_type="jesuita-entrada",
        the_value="Coimbra",
        show_elements=["name", "sex", "extra_info"],
        more_attributes=["nascimento"],
        sql_echo=True,
    )
    assert df is not None, "entities_w,ith_attribute returned None"
    assert len(df) > 0, "entities_with_attribute returned empty dataframe"
    print(df)


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_entities_with_attribute_filter_by(dbsystem):
    """Test generation of dataframe from attributes"""
    conimbricensis = [
        "deh-afonso-aires",
        "deh-antonio-de-abreu-ref1",
        "deh-antonio-de-andrade",
        "deh-goncalo-alvares",
        "deh-miguel-do-amaral",
        "deh-pedro-de-alcacova",
    ]
    df = entities_with_attribute(
        entity_type="person",
        show_elements=["name", "groupname"],
        the_type="jesuita-entrada",
        more_attributes=["nacionalidade", "nascimento"],
        filter_by=conimbricensis,
        sql_echo=True,
        db=dbsystem,
    )
    assert df is not None, "entities_with_attribute returned None"
    assert len(df.index.unique()) == len(
        conimbricensis
    ), "entities_with_attribute returned bad filtered dataframe"
    print(df)


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_entities_with_attribute_list(dbsystem):
    """Test generation of dataframe from attributes"""
    df = entities_with_attribute(
        db=dbsystem,
        entity_type="person",
        the_type=[
            "nascimento",
            "entrada",
            "partida",
            "estadia",
            "morte",
            "jesuita-entrada",
            "jesuita-votos-local",
        ],
        column_name="lugares",
        show_elements=["name", "sex", "the_line"],
        more_attributes=[],
        filter_by=["deh-ludovico-antonio-adorno"],
        sql_echo=True,
    )
    # pdb.set_trace()
    assert df is not None, "entities_with_attribute_list returned None"
    assert len(df) > 0, "entities_with_attribute_list returned empty dataframe"
    df_cols = list(df.columns)
    required_column = "lugares.original"
    assert required_column in df_cols, f"{required_column} expected in {df_cols}"
    required_column = "lugares.comment"
    assert required_column in df_cols, f"{required_column} expected in {df_cols}"
    print(df.info())


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_entities_with_attribute_empty(dbsystem):
    """Test generation of dataframe from attributes"""
    df = entities_with_attribute(
        db=dbsystem,
        entity_type="person",
        the_type="nonexistent",
        column_name="lugares",
        show_elements=["name", "sex", "the_line"],
        more_attributes=[],
        filter_by=["deh-ludovico-antonio-adorno"],
        sql_echo=True,
    )
    assert df is None, "entities_with_attribute failed to return None"

    df = entities_with_attribute(
        db=dbsystem,
        entity_type="person",
        the_type="nascimento",
        column_name="lugares",
        show_elements=["name", "sex", "the_line"],
        more_attributes=["nonexistent"],
        filter_by=["deh-ludovico-antonio-adorno"],
        sql_echo=True,
    )
    assert df is not None, "entities_with_attribute returned None"



# flake8: noqa: F401
from typing import List
from .kleio_server import KleioServer
from .schemas import KleioFile
from .schemas import ApiPermissions, TokenInfo
from .schemas import translation_status_enum
from .schemas import import_status_enum

api_permissions_normal: List[ApiPermissions] = [
    ApiPermissions.sources,
    ApiPermissions.kleioset,
    ApiPermissions.files,
    ApiPermissions.structures,
    ApiPermissions.translations,
    ApiPermissions.upload,
    ApiPermissions.delete,
    ApiPermissions.mkdir,
    ApiPermissions.rmdir,
]

token_info_normal: TokenInfo = TokenInfo(
    comment="An user able to translate, upload and delete files, and also create and remove directories, in specific sub-directoris in kleio-home",
    api=api_permissions_normal,
    structures="structures/reference_sources",
    sources="sources/reference_sources",
)



"""Add blinks table for persistent rids

Revision ID: 36cdd078f890
Revises: 48dd68d06c60
Create Date: 2024-11-30 21:00:23.413029

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "36cdd078f890"
down_revision: Union[str, None] = "48dd68d06c60"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    # check if table exists
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    tables = inspector.get_table_names()
    if "blinks" not in tables:
        op.create_table(
            "blinks",
            sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
            sa.Column("rid", sa.String(length=64), nullable=False),
            sa.Column("entity", sa.String(length=64), nullable=True),
            sa.Column("user", sa.String(length=64), nullable=False),
            sa.Column("source", sa.String(length=64), nullable=True),
            sa.Column("rule", sa.String(length=4096), nullable=False),
            sa.Column(
                "status",
                sa.Enum(
                    "UNLINKED",
                    "AUTOMATIC",
                    "SOURCE",
                    "MANUAL",
                    "MHK",
                    "N",
                    "INVALID",
                    name="linkstatus",
                ),
                nullable=False,
            ),
            sa.Column("aregister", sa.String(length=64), nullable=True),
            sa.ForeignKeyConstraint(["rid"], ["rentities.id"], ondelete="CASCADE"),
            sa.PrimaryKeyConstraint("id"),
        )
        op.create_index(op.f("ix_blinks_entity"), "blinks", ["entity"], unique=False)
        op.create_index(op.f("ix_blinks_rid"), "blinks", ["rid"], unique=False)
        op.create_index(op.f("ix_blinks_source"), "blinks", ["source"], unique=False)

    with op.batch_alter_table("links") as batch_op:
        try:
            batch_op.drop_constraint(None, type_="foreignkey")
        except Exception as e:
            print(f"Caught Error dropping constraint: {e}")
        fks = inspector.get_foreign_keys(table_name='links')
        for fk in fks:
            if fk['referred_table'] == 'entities':
                try:
                    batch_op.drop_constraint(fk['name'], type_="foreignkey")
                except Exception as e:
                    print(f"Caught Error dropping constraint: {e}")
        batch_op.create_foreign_key(None, "entities", ["entity"], ["id"])
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    # this is a workaround for sqlite lack of drop_constrant support
    with op.batch_alter_table("links") as batch_op:
        # cant make the drop_constraint work
        # batch_op.drop(None, type_="foreignkey")
        batch_op.create_foreign_key(
            None,
            referent_table="entities",
            local_cols=["entity"],
            remote_cols=["id"],
            ondelete="SET NULL"
        )

    op.drop_index(op.f("ix_blinks_source"), table_name="blinks")
    op.drop_index(op.f("ix_blinks_rid"), table_name="blinks")
    op.drop_index(op.f("ix_blinks_entity"), table_name="blinks")
    op.drop_table("blinks")
    # ### end Alembic commands ###



# This is necessary to avoid error in docs generation
# File "/Users/jrc/develop/timelink-py/timelink/app/main.py", line 133, in <module>
#     app.mount("/static", StaticFiles(packages=[("timelink")]), name="static")
#                          ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#   File "/Users/jrc/develop/timelink-py/.venv/lib/python3.11/site-packages/starlette/staticfiles.py", line 54, in __init__
#     self.all_directories = self.get_directories(directory, packages)
#                            ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#   File "/Users/jrc/develop/timelink-py/.venv/lib/python3.11/site-packages/starlette/staticfiles.py", line 85, in get_directories
#     assert os.path.isdir(
# AssertionError: Directory ''statics'' in package 'timelink' could not be found.



"""
(c) Joaquim Carvalho 2023.
MIT License, no warranties.
"""

from typing import Optional

from sqlalchemy import String, ForeignKey
from sqlalchemy.orm import Mapped  # pylint: disable=import-error
from sqlalchemy.orm import mapped_column  # pylint: disable=import-error

from .entity import Entity
from timelink.kleio.utilities import kleio_escape, quote_long_text


class Source(Entity):
    """Represent an historical source"""

    __tablename__ = "sources"

    id: Mapped[str] = mapped_column(
        String, ForeignKey("entities.id", ondelete="CASCADE"), primary_key=True
    )
    the_type: Mapped[Optional[str]] = mapped_column(
        String(32),
        nullable=True,
        comment="type of the source (parish register, tax roll, etc.)",
    )
    the_date: Mapped[Optional[str]] = mapped_column(String, nullable=True, index=True)
    loc: Mapped[Optional[str]] = mapped_column(
        String, nullable=True, comment="location of the source"
    )
    ref: Mapped[Optional[str]] = mapped_column(
        String, nullable=True, comment="call number of the source"
    )
    kleiofile: Mapped[Optional[str]] = mapped_column(
        String, index=True, comment="path of the kleio file"
    )
    replaces: Mapped[Optional[str]] = mapped_column(
        String, nullable=True, comment="id of the source this one replaces"
    )
    obs: Mapped[Optional[str]] = mapped_column(
        String, nullable=True, comment="observations"
    )

    __mapper_args__ = {"polymorphic_identity": "source"}

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Source(id={sr}, "
            f'the_type="{self.the_type}", '
            f'the_date="{self.the_date}", '
            f'local="{self.loc}", '
            f'ref="{self.ref}", '
            f'kleiofile="{self.kleiofile}", '
            f'replaces="{self.replaces}", '
            f'obs="{self.obs}"'
            f")"
        )

    def __str__(self):
        return self.to_kleio()

    def to_kleio(
        self, ident="", ident_inc="  ", show_contained=True, width=80, **kwargs
    ) -> str:
        r = (
            f"{self.groupname}${self.id}/{self.the_date}"
            f"/type={kleio_escape(self.the_type)}"
            f"/ref={kleio_escape(self.ref)}"
            f"/loc={kleio_escape(self.loc)}"
            f'/kleiofile={kleio_escape(self.kleiofile)}"'
            f"/replaces={self.replaces}"
        )
        if self.obs is not None and len(self.obs.strip()) > 0:
            r = f"{r}/obs={quote_long_text(self.obs.strip(), width=width)}"
        r = super().to_kleio(
            self_string=r,
            ident=ident,
            ident_inc=ident_inc,
            show_contained=show_contained,
            width=width,
            **kwargs
        )
        return r
